import datetime
import json
import logging

import mock
import numpy as np
import pandas as pd
import pytest
import responses
from freezegun import freeze_time

from gantry.exceptions import GantryBatchCreationException, GantryLoggingDataTypeError
from gantry.logger import client
from gantry.logger.constants import BatchType, UploadFileType
from gantry.logger.stores import BaseLogStore
from gantry.logger.types import DataLink, DataLinkElement

CURRENT_TIME, CURRENT_TIME_STR = datetime.datetime(2032, 5, 23, 0, 0, 0), "2032-05-23T00:00:00"
FUTURE_TIME = datetime.datetime(2042, 5, 23, 0, 0, 0)
SOME_TIME, SOME_TIME_STR = datetime.datetime(2022, 5, 10, 17, 50, 32), "2022-05-10T17:50:32"
ANOTHER_TIME, ANOTHER_TIME_STR = datetime.datetime(2022, 5, 11, 17, 50, 32), "2022-05-11T17:50:32"
HOST = "http://test-api"


class TestLogStore(BaseLogStore):
    def __init__(self):
        self.logs = []
        self.status = False

    def log_batch(self, application, events) -> None:
        self.logs.append({"application": application, "events": events})

    def ping(self) -> bool:
        return self.status


@pytest.fixture(scope="function")
def log_store():
    return TestLogStore()


@pytest.fixture(scope="function")
def cli_obj(log_store):
    return client.Gantry(log_store=log_store, environment="test", logging_level="DEBUG")


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize(
    ["sort", "timestamps", "record_count", "expected"],
    [
        (False, [3, 4, 5, 1, 2], 10, [(0, 3), (1, 4), (2, 5), (3, 1), (4, 2)]),
        (True, [3, 4, 5, 1, 2], 10, [(3, 1), (4, 2), (0, 3), (1, 4), (2, 5)]),
        (False, ["b", "a", "c"], 100, [(0, "b"), (1, "a"), (2, "c")]),
        (True, ["b", "a", "c"], 100, [(1, "a"), (0, "b"), (2, "c")]),
        (
            True,
            None,
            4,
            [
                (0, CURRENT_TIME),
                (1, CURRENT_TIME),
                (2, CURRENT_TIME),
                (3, CURRENT_TIME),
            ],
        ),
        (
            False,
            None,
            4,
            [
                (0, CURRENT_TIME),
                (1, CURRENT_TIME),
                (2, CURRENT_TIME),
                (3, CURRENT_TIME),
            ],
        ),
    ],
)
def test_create_timestamp_idx(sort, timestamps, record_count, expected):
    assert list(client._create_timestamp_idx(sort, timestamps, record_count)) == expected


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize("feedback_version", [None, 5, "1.2.3"])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_build_feedback_event(mock_uuid, feedback_version, timestamp):
    mock_uuid.return_value = "123abc"
    assert client._build_feedback_event(
        context={"foo": "bar"},
        application="barbaz",
        feedback_id={"a": 10},
        feedback={"b": 20},
        feedback_version=feedback_version,
        timestamp=timestamp,
        batch_id="ABCD1234",
    ) == {
        "batch_id": "ABCD1234",
        "event_id": "123abc",
        "feedback": {"b": 20},
        "feedback_id": "2a30f5f3b7d1a97cb6132480b992d984",
        "feedback_id_inputs": {"a": 10},
        "log_timestamp": CURRENT_TIME_STR,
        "metadata": {"feedback_version": feedback_version, "foo": "bar", "func_name": "barbaz"},
        "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
    }


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize(
    ["feedback_keys", "feedback_id"],
    [
        (None, None),
        (["some-input"], None),
        (None, {"id": "yes"}),
    ],
)
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_build_prediction_event(mock_uuid, timestamp, feedback_keys, feedback_id):
    mock_uuid.return_value = "123abc"
    assert client._build_prediction_event(
        context={"foo": "bar"},
        env="test",
        inputs={"some-input": "a-input"},
        outputs={"some-output": "a-output"},
        application="barbaz",
        version="1.2.3",
        feedback_keys=feedback_keys,
        ignore_inputs=["some-other-input"],
        batch_id="ABCD1234",
        feedback_id=feedback_id,
        tags={"this-is": "a-tag"},
        custom_timestamp=timestamp,
    ) == {
        "batch_id": "ABCD1234",
        "event_id": "123abc",
        "feedback_id": "1f972d1df351de3ce35a787c89faad29"
        if feedback_id
        else "afd8cf8889da593bcb98dae5c86101ca",
        "inputs": {"some-input": "a-input"},
        "log_timestamp": CURRENT_TIME_STR,
        "metadata": {
            "feedback_keys": feedback_keys,
            "foo": "bar",
            "version": "1.2.3",
            "ignore_inputs": ["some-other-input"],
            "provided_feedback_id": feedback_id,
            "func_name": "barbaz",
        },
        "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
        "outputs": {"some-output": "a-output"},
        "tags": {"env": "test", "this-is": "a-tag"},
    }

    with pytest.raises(GantryLoggingDataTypeError):
        client._build_prediction_event(
            context={"foo": "bar"},
            env="test",
            inputs={"some-invalid-input": ["invalid", {"invalid": True}]},
            outputs={"some-output": "a-output"},
            application="barbaz",
            version="1.2.3",
            feedback_keys=[],
            ignore_inputs=[],
            batch_id="ABCD1234",
            feedback_id=feedback_id,
            tags={"this-is": "a-tag"},
            custom_timestamp=timestamp,
        )


def test_ping_fail(cli_obj):
    cli_obj.log_store.status = False
    assert not cli_obj.ping()


def test_ping_success(cli_obj):
    cli_obj.log_store.status = True
    assert cli_obj.ping()


@pytest.mark.parametrize("application", [10, None, 10.5, b"some binary string"])
def test_log_feedback_event_invalid_name(application, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application=application,
        feedback_id={"A": 100},
        feedback={"value": "bar"},
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_id", [10, None, 10.5, ["a", "list"], b"something"])
def test_log_feedback_event_invalid_feedback_id(feedback_id, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id=feedback_id,
        feedback={"value": "bar"},
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback", [10, None, 10.5, ["a", "list"], b"something"])
def test_log_feedback_event_invalid_feedback(feedback, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback=feedback,
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_version", [b"a-version", 10.5, {"version": "yes"}, [10]])
def test_log_feedback_event_invalid_feedback_version(feedback_version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback={"value": "foo"},
        feedback_version=feedback_version,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("timestamp", ["some-timestamp", 10, "2020-10-10"])
def test_log_feedback_event_invalid_timestamp(timestamp, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback={"value": "foo"},
        feedback_version=10,
        timestamp=timestamp,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("timestamp", [FUTURE_TIME])
def test_log_feedback_event_future_timestamp(timestamp, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback={"value": "foo"},
        feedback_version=10,
        timestamp=timestamp,
    )

    # Assert that there were no logs written, since the dates are in the future.
    assert len(cli_obj.log_store.logs) == 0


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize("feedback_version", [None, "1.2.3", 10])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_event(mock_uuid, feedback_version, timestamp, tags, cli_obj):
    mock_uuid.return_value = "12345"
    cli_obj.log_feedback_event(
        application="foobar",
        feedback_id={"A": 100},
        feedback={"value": "bar"},
        feedback_version=feedback_version,
        timestamp=timestamp,
        tags=tags,
    )
    assert cli_obj.log_store.logs == [
        {
            "events": [
                {
                    "batch_id": None,
                    "event_id": "12345",
                    "feedback": {"value": "bar"},
                    "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
                    "feedback_id_inputs": {"A": 100},
                    "log_timestamp": CURRENT_TIME_STR,
                    "metadata": {
                        "environment": tags["env"] if tags else "test",
                        "feedback_version": feedback_version,
                        "func_name": "foobar",
                    },
                    "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
                }
            ],
            "application": "foobar",
        },
    ]


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_event_sample_rate_0_5(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        gantry.log_feedback_event(
            application="foobar",
            feedback_id={"A": 100},
            feedback={"value": "bar"},
            feedback_version=10,
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_called_once()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_event_sample_rate_0_5_none(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")
    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "12345"
        gantry.log_feedback_event(
            application="foobar",
            feedback_id={"A": 100},
            feedback={"value": "bar"},
            feedback_version=10,
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_event_sample_rate_0(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        mock_uuid.return_value = "12345"
        gantry.log_feedback_event(
            application="foobar",
            feedback_id={"A": 100},
            feedback={"value": "bar"},
            feedback_version=10,
            timestamp=timestamp,
            sample_rate=0.0,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("application", [10, None, 10.5, b"some binary string"])
def test_log_prediction_event_invalid_name(application, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application=application,
        inputs={"A": 100},
        outputs={"value": "bar"},
        version=10,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("inputs", [[10, 20], 10, None, 10.5, ("A", 100)])
def test_log_prediction_event_invalid_inputs(inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs=inputs,
        outputs={"value": "bar"},
        version=10,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("version", [[10, 20], 10.5, {"version": 10}])
def test_log_prediction_event_invalid_version(version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version=version,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_keys", [[10, 20], 10.5, ("a", "b", "c")])
def test_log_prediction_event_invalid_feedback_keys(feedback_keys, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=feedback_keys,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize(
    "ignore_inputs", [[10, 20], [b"something", "binary"], ["bar", 10.5], [10, "foo"], 10, 10.5]
)
def test_log_prediction_event_invalid_ignore_inputs(ignore_inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        ignore_inputs=ignore_inputs,
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_id", [[10, 20], ["foo"], 10, 10.5])
def test_log_prediction_event_invalid_feedback_id(feedback_id, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=["A"],
        ignore_inputs=["A"],
        feedback_id=feedback_id,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("timestamp", [[], "some-timestamp", 10, "2020-10-10"])
def test_log_prediction_event_invalid_timestamp(timestamp, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_prediction_event(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=["A"],
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=timestamp,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize(
    ["feedback_keys", "feedback_id"],
    [
        (["A"], None),
        (None, {"A": 100}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_event(mock_uuid, feedback_keys, feedback_id, timestamp, tags, cli_obj):
    mock_uuid.return_value = "12345"
    cli_obj.log_prediction_event(
        application="foobar",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=feedback_keys,
        ignore_inputs=["A"],
        feedback_id=feedback_id,
        timestamp=timestamp,
        tags=tags,
    )
    assert cli_obj.log_store.logs == [
        {
            "events": [
                {
                    "batch_id": None,
                    "event_id": "12345",
                    "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
                    "inputs": {},
                    "log_timestamp": CURRENT_TIME_STR,
                    "metadata": {
                        "environment": tags["env"] if tags else "test",
                        "feedback_keys": ["A"] if feedback_keys else None,
                        "func_name": "foobar",
                        "ignore_inputs": ["A"],
                        "provided_feedback_id": {"A": 100} if feedback_id else None,
                        "version": "a.b.c",
                    },
                    "outputs": {"value": "bar"},
                    "tags": tags if tags else {"env": "test"},
                    "timestamp": CURRENT_TIME_STR if not timestamp else SOME_TIME_STR,
                }
            ],
            "application": "foobar",
        },
    ]


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize(
    ["feedback_keys", "feedback_id"],
    [
        (["A"], None),
        (None, {"A": 100}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_event_sample_rate_0_5(mock_uuid, feedback_keys, feedback_id, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        mock_uuid.return_value = "12345"
        gantry.log_prediction_event(
            application="foobar",
            inputs={"A": 100},
            outputs={"value": "bar"},
            version="a.b.c",
            feedback_keys=feedback_keys,
            ignore_inputs=["A"],
            feedback_id=feedback_id,
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_called_once()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize(
    ["feedback_keys", "feedback_id"],
    [
        (["A"], None),
        (None, {"A": 100}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_event_sample_rate_0_5_none(
    mock_uuid, feedback_keys, feedback_id, timestamp
):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "12345"
        gantry.log_prediction_event(
            application="foobar",
            inputs={"A": 100},
            outputs={"value": "bar"},
            version="a.b.c",
            feedback_keys=feedback_keys,
            ignore_inputs=["A"],
            feedback_id=feedback_id,
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@pytest.mark.parametrize(
    ["feedback_keys", "feedback_id"],
    [
        (["A"], None),
        (None, {"A": 100}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_event_sample_rate_0(mock_uuid, feedback_keys, feedback_id, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "12345"
        gantry.log_prediction_event(
            application="foobar",
            inputs={"A": 100},
            outputs={"value": "bar"},
            version="a.b.c",
            feedback_keys=feedback_keys,
            ignore_inputs=["A"],
            feedback_id=feedback_id,
            timestamp=timestamp,
            sample_rate=0.0,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("application", [10, None, 10.5, b"some binary string"])
def test_log_record_invalid_name(application, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application=application,
        feedback_id={"A": 100},
        feedback={"value": "bar"},
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_id", [10, None, 10.5, ["a", "list"], b"something"])
def test_log_record_invalid_feedback_id(feedback_id, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="foobar",
        feedback_id=feedback_id,
        feedback={"value": "bar"},
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback", [10, None, 10.5, ["a", "list"], b"something"])
def test_log_record_invalid_feedback(feedback, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback=feedback,
        feedback_version=10,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize(
    "feedback_version", ["some-version", b"a-version", 10.5, {"version": "yes"}, [10]]
)
def test_log_record_invalid_feedback_version(feedback_version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback={"value": "foo"},
        feedback_version=feedback_version,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("timestamp", ["some-timestamp", 10, "2020-10-10"])
def test_log_record_invalid_timestamp(timestamp, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="foobar",
        feedback_id={"value": "bar"},
        feedback={"value": "foo"},
        feedback_version=10,
        timestamp=timestamp,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("inputs", [[10, 20], 10, 10.5, ("A", 100)])
def test_log_record_invalid_inputs(inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs=inputs,
        outputs={"value": "bar"},
        version=10,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("version", [[10, 20], 10.5, {"version": 10}])
def test_log_record_invalid_version(version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version=version,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_keys", [[10, 20], 10.5, ("a", "b", "c")])
def test_log_record_invalid_feedback_keys(feedback_keys, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=feedback_keys,
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize(
    "ignore_inputs", [[10, 20], [b"something", "binary"], ["bar", 10.5], [10, "foo"], 10, 10.5]
)
def test_log_record_invalid_ignore_inputs(ignore_inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        ignore_inputs=ignore_inputs,
        feedback_id={"A": 100},
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_id", [[10, 20], ["foo"], 10, 10.5])
def test_log_record_invalid_feedback_id_for_prediction(feedback_id, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=["A"],
        ignore_inputs=["A"],
        feedback_id=feedback_id,
        timestamp=SOME_TIME,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("feedback_id", ["foo", {"foo": "bar"}])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_record_valid_feedback_id_for_prediction(mock_uuid, feedback_id, cli_obj):
    mock_uuid.return_value = "12345"
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        ignore_inputs=["B"],
        feedback_id=feedback_id,
        timestamp=SOME_TIME,
    )
    assert cli_obj.log_store.logs == [
        {
            "events": [
                {
                    "batch_id": None,
                    "event_id": "12345",
                    "inputs": {"A": 100},
                    "outputs": {"value": "bar"},
                    "tags": {"env": "test"},
                    "feedback_id": "42a5b7b99b3717d1eaeb72a6948dabc9"
                    if isinstance(feedback_id, dict)
                    else feedback_id,
                    "log_timestamp": CURRENT_TIME_STR,
                    "metadata": {
                        "environment": "test",
                        "func_name": "some-name",
                        "provided_feedback_id": feedback_id,
                        "version": "a.b.c",
                        "ignore_inputs": ["B"],
                        "feedback_keys": None,
                    },
                    "timestamp": SOME_TIME_STR,
                }
            ],
            "application": "some-name",
        },
    ]


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("feedback_id", ["foo", {"foo": "bar"}])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_record_valid_feedback_id_for_feedback(mock_uuid, feedback_id, cli_obj):
    mock_uuid.return_value = "12345"
    cli_obj.log_record(
        application="some-name",
        feedback={"A": 100},
        version="a.b.c",
        feedback_id=feedback_id,
        timestamp=SOME_TIME,
    )
    assert cli_obj.log_store.logs == [
        {
            "events": [
                {
                    "batch_id": None,
                    "event_id": "12345",
                    "feedback": {"A": 100},
                    "feedback_id": "42a5b7b99b3717d1eaeb72a6948dabc9"
                    if isinstance(feedback_id, dict)
                    else feedback_id,
                    "feedback_id_inputs": feedback_id,
                    "log_timestamp": CURRENT_TIME_STR,
                    "metadata": {
                        "environment": "test",
                        "func_name": "some-name",
                        "feedback_version": "a.b.c",
                    },
                    "timestamp": SOME_TIME_STR,
                }
            ],
            "application": "some-name",
        },
    ]


@pytest.mark.parametrize("timestamp", [[], "some-timestamp", 10, "2020-10-10"])
def test_log_record_invalid_timestamp_for_prediction(timestamp, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_record(
        application="some-name",
        inputs={"A": 100},
        outputs={"value": "bar"},
        version="a.b.c",
        feedback_keys=["A"],
        ignore_inputs=["A"],
        feedback_id={"A": 100},
        timestamp=timestamp,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_and_feedback_event(mock_uuid, timestamp, tags, cli_obj):
    mock_uuid.return_value = "12345"
    cli_obj._log_prediction_and_feedback_event(
        application="foobar",
        version="a.b.c",
        inputs={"A": 100},
        outputs={"value": "bar"},
        feedback={"value": "potato"},
        timestamp=timestamp,
        tags=tags,
    )

    assert cli_obj.log_store.logs == [
        {
            "events": [
                {
                    "batch_id": None,
                    "event_id": "12345",
                    "feedback": {"value": "potato"},
                    "inputs": {"A": 100},
                    "outputs": {"value": "bar"},
                    "tags": tags if tags else {"env": "test"},
                    "feedback_id": "d751713988987e9331980363e24189ce",
                    "feedback_id_inputs": {},
                    "log_timestamp": CURRENT_TIME_STR,
                    "metadata": {
                        "environment": tags["env"] if tags else "test",
                        "feedback_version": "a.b.c",
                        "func_name": "foobar",
                        "provided_feedback_id": None,
                        "version": "a.b.c",
                        "ignore_inputs": None,
                        "feedback_keys": None,
                    },
                    "timestamp": SOME_TIME_STR if timestamp else CURRENT_TIME_STR,
                }
            ],
            "application": "foobar",
        },
    ]


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_and_feedback_event_sample_rate_0_5(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        mock_uuid.return_value = "12345"
        gantry._log_prediction_and_feedback_event(
            application="foobar",
            version="a.b.c",
            inputs={"A": 100},
            outputs={"value": "bar"},
            feedback={"value": "potato"},
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_called_once()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_and_feedback_event_sample_rate_0_5_none(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "12345"
        gantry._log_prediction_and_feedback_event(
            application="foobar",
            version="a.b.c",
            inputs={"A": 100},
            outputs={"value": "bar"},
            feedback={"value": "potato"},
            timestamp=timestamp,
            sample_rate=0.5,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("timestamp", [None, SOME_TIME])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_prediction_and_feedback_event_sample_rate_0(mock_uuid, timestamp):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "12345"
        gantry._log_prediction_and_feedback_event(
            application="foobar",
            version="a.b.c",
            inputs={"A": 100},
            outputs={"value": "bar"},
            feedback={"value": "potato"},
            timestamp=timestamp,
            sample_rate=0.0,
        )

    log_store.log.assert_not_called()


@pytest.mark.parametrize("application", [10, None, 10.5, b"some binary string"])
def test_log_feedback_invalid_name(application, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback(
        application=application,
        feedback_ids=[{"A": 100}],
        feedbacks=[{"value": "bar"}],
        feedback_version=10,
        timestamps=[SOME_TIME],
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_ids", [[10], 10, None, 10.5])
def test_log_feedback_invalid_feedback_id(feedback_ids, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback(
        application="foobar",
        feedback_ids=feedback_ids,
        feedbacks=[{"value": "bar"}],
        feedback_version=10,
        timestamps=[SOME_TIME],
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedbacks", [10, None, 10.5, np.array([1, 2])])
def test_log_feedback_invalid_feedback(feedbacks, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback(
        application="foobar",
        feedback_ids=[{"value": "bar"}],
        feedbacks=feedbacks,
        feedback_version=10,
        timestamps=[SOME_TIME],
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_version", [{"this-is": 10}, ["1.2.3", "2.3.4"]])
def test_log_feedback_invalid_feedback_version(feedback_version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback(
        application="foobar",
        feedback_ids=[{"value": "bar"}],
        feedbacks=[{"value": "foo"}],
        feedback_version=feedback_version,
        timestamps=[SOME_TIME],
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("timestamps", [[10], "some-timestamp", 10, ["2020-20-20", SOME_TIME]])
def test_log_feedback_invalid_timestamps(timestamps, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_feedback(
        application="foobar",
        feedback_ids=[{"value": "bar"}],
        feedbacks=[{"value": "foo"}],
        feedback_version=10,
        timestamps=timestamps,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


# TODO -> add np.ndarrays as possible feedbacks values
@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize("random_error", [True, False])
@pytest.mark.parametrize(
    "feedbacks",
    [
        [{100: "some-value"}, {100: "some-other-value"}],
        pd.DataFrame({100: ["some-value", "some-other-value"]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback(mock_uuid, feedbacks, random_error, timestamps, sort, tags, cli_obj):
    if random_error:
        mock_uuid.side_effect = RuntimeError
    else:
        mock_uuid.return_value = "ABCD1234"
    rv = cli_obj.log_feedback(
        application="foobar",
        feedback_ids=[{"A": 100}, {"B": 200}],
        feedbacks=feedbacks,
        feedback_version=10,
        timestamps=timestamps,
        sort_on_timestamp=sort,
        as_batch=False,
        tags=tags,
    )
    assert rv is None
    events = [
        {
            "batch_id": None,
            "event_id": "ABCD1234",
            "feedback": {100: "some-other-value"},
            "feedback_id": "08b84280663845e83bbf82d297013a53",
            "feedback_id_inputs": {"B": 200},
            "log_timestamp": CURRENT_TIME_STR,
            "metadata": {
                "environment": tags["env"] if tags else "test",
                "feedback_version": 10,
                "func_name": "foobar",
            },
            "timestamp": SOME_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        },
        {
            "batch_id": None,
            "event_id": "ABCD1234",
            "feedback": {100: "some-value"},
            "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
            "feedback_id_inputs": {"A": 100},
            "log_timestamp": CURRENT_TIME_STR,
            "metadata": {
                "environment": tags["env"] if tags else "test",
                "feedback_version": 10,
                "func_name": "foobar",
            },
            "timestamp": ANOTHER_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        },
    ]

    if timestamps is not None:
        expected_events = [events[1], events[0]] if not sort else [events[0], events[1]]
    else:
        expected_events = [events[1], events[0]]

    if random_error:
        assert cli_obj.log_store.logs == []
    else:
        assert cli_obj.log_store.logs == [
            {
                "events": expected_events,
                "application": "foobar",
            }
        ]


@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "feedbacks",
    [
        [{100: "some-value"}, {100: "some-other-value"}],
        pd.DataFrame({100: ["some-value", "some-other-value"]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_sample_rate_0_5(mock_uuid, feedbacks, timestamps, sort):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_feedback(
            application="foobar",
            feedback_ids=[{"A": 100}, {"B": 200}],
            feedbacks=feedbacks,
            feedback_version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.5,
            as_batch=False,
        )

    log_store.log_batch.assert_called_once()


@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "feedbacks",
    [
        [{100: "some-value"}, {100: "some-other-value"}],
        pd.DataFrame({100: ["some-value", "some-other-value"]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_sample_rate_0_5_none(mock_uuid, feedbacks, timestamps, sort, cli_obj):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_feedback(
            application="foobar",
            feedback_ids=[{"A": 100}, {"B": 200}],
            feedbacks=feedbacks,
            feedback_version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.5,
            as_batch=False,
        )

    log_store.log_batch.assert_not_called()


@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "feedbacks",
    [
        [{100: "some-value"}, {100: "some-other-value"}],
        pd.DataFrame({100: ["some-value", "some-other-value"]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_feedback_sample_rate_0_0(mock_uuid, feedbacks, timestamps, sort):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_feedback(
            application="foobar",
            feedback_ids=[{"A": 100}, {"B": 200}],
            feedbacks=feedbacks,
            feedback_version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.0,
            as_batch=False,
        )

    log_store.log_batch.assert_not_called()


def test_log_feedback_mismatch_size(cli_obj):
    cli_obj.log_feedback(
        application="foobar",
        feedback_ids=[{"A": 100}],
        feedbacks=[{100: "some-value"}, {200: "some-other-value"}],
        feedback_version=10,
        timestamps=[ANOTHER_TIME, SOME_TIME],
        sort_on_timestamp=True,
        as_batch=False,
    )
    assert cli_obj.log_store.logs == []


@pytest.mark.parametrize("application", [10, None, 10.5, b"some binary string"])
def test_log_predictions_invalid_name(application, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application=application,
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["A", "B"],
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize(
    "inputs", [[10, {}], {"A": 100}, 10, None, 10.5, b"some binary string", np.array([1, 2])]
)
def test_log_predictions_invalid_inputs(inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=inputs,
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["A", "B"],
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize(
    "outputs", [{"A": 100}, 10, None, 10.5, b"some binary string", np.array([1, 2])]
)
def test_log_predictions_invalid_outputs(outputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=outputs,
        feedback_keys=["A"],
        ignore_inputs=["A", "B"],
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("version", [{"A": 100}, 10.5, b"some binary string"])
def test_log_predictions_invalid_version(version, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["A", "B"],
        version=version,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_keys", [[{}], {"A": 100}, 10.5, b"some binary string"])
def test_log_predictions_invalid_feedback_keys(feedback_keys, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=feedback_keys,
        ignore_inputs=["A", "B"],
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("ignore_inputs", [[{}], {"A": 100}, 10.5, b"some binary string"])
def test_log_predictions_invalid_ignore_inputs(ignore_inputs, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=ignore_inputs,
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("feedback_ids", [[10], {"A": 100}, 10.5, b"some binary string"])
def test_log_predictions_invalid_feedback_ids(feedback_ids, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        ignore_inputs=["A", "B"],
        feedback_ids=feedback_ids,
        version=10,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


@pytest.mark.parametrize("timestamps", [[10], "some-timestamp", 10, ["2020-20-20", SOME_TIME]])
def test_log_predictions_invalid_timestamps(timestamps, cli_obj, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["A", "B"],
        version=10,
        timestamps=timestamps,
        as_batch=False,
    )
    assert "Internal exception in Gantry client" in caplog.text
    assert "TypeError:" in caplog.text


# TODO -> add np.ndarrays as possible inputs/outputs values
@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize("random_error", [True, False])
@pytest.mark.parametrize(
    "outputs",
    [
        [{"C": 300}, {"C": 303}],
        pd.DataFrame({"C": [300, 303]}),
    ],
)
@pytest.mark.parametrize(
    "inputs",
    [
        [{"A": 100, "B": 200}, {"A": 101, "B": 202}],
        pd.DataFrame({"A": [100, 101], "B": [200, 202]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_predictions(mock_uuid, inputs, outputs, random_error, timestamps, sort, tags, cli_obj):
    if random_error:
        mock_uuid.side_effect = RuntimeError
    else:
        mock_uuid.return_value = "ABCD1234"
    rv = cli_obj.log_predictions(
        application="foobar",
        inputs=inputs,
        outputs=outputs,
        feedback_keys=["A"],
        ignore_inputs=["B"],
        version=10,
        timestamps=timestamps,
        sort_on_timestamp=sort,
        as_batch=False,
        tags=tags,
    )
    assert rv is None

    events = [
        {
            "batch_id": None,
            "event_id": "ABCD1234",
            "feedback_id": "9e329293e022d6cdaafdec49b5f4fedc",
            "inputs": {"A": 101},
            "log_timestamp": CURRENT_TIME_STR,
            "metadata": {
                "environment": tags["env"] if tags else "test",
                "feedback_keys": ["A"],
                "func_name": "foobar",
                "ignore_inputs": ["B"],
                "provided_feedback_id": None,
                "version": 10,
            },
            "outputs": {"C": 303},
            "tags": tags if tags else {"env": "test"},
            "timestamp": SOME_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        },
        {
            "batch_id": None,
            "event_id": "ABCD1234",
            "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
            "inputs": {"A": 100},
            "log_timestamp": CURRENT_TIME_STR,
            "metadata": {
                "environment": tags["env"] if tags else "test",
                "feedback_keys": ["A"],
                "func_name": "foobar",
                "ignore_inputs": ["B"],
                "provided_feedback_id": None,
                "version": 10,
            },
            "outputs": {"C": 300},
            "tags": tags if tags else {"env": "test"},
            "timestamp": ANOTHER_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        },
    ]

    if timestamps is not None:
        expected_events = [events[1], events[0]] if not sort else [events[0], events[1]]
    else:
        expected_events = [events[1], events[0]]

    if random_error:
        assert cli_obj.log_store.logs == []
    else:
        assert cli_obj.log_store.logs == [
            {
                "events": expected_events,
                "application": "foobar",
            }
        ]


@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "outputs",
    [
        [{"C": 300}, {"C": 303}],
        pd.DataFrame({"C": [300, 303]}),
    ],
)
@pytest.mark.parametrize(
    "inputs",
    [
        [{"A": 100, "B": 200}, {"A": 101, "B": 202}],
        pd.DataFrame({"A": [100, 101], "B": [200, 202]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_predictions_sample_rate_0_5(mock_uuid, inputs, outputs, timestamps, sort, tags):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_predictions(
            application="foobar",
            inputs=inputs,
            outputs=outputs,
            feedback_keys=["A"],
            ignore_inputs=["B"],
            version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.5,
            as_batch=False,
            tags=tags,
        )
    log_store.log_batch.assert_called_once()


@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "outputs",
    [
        [{"C": 300}, {"C": 303}],
        pd.DataFrame({"C": [300, 303]}),
    ],
)
@pytest.mark.parametrize(
    "inputs",
    [
        [{"A": 100, "B": 200}, {"A": 101, "B": 202}],
        pd.DataFrame({"A": [100, 101], "B": [200, 202]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_predictions_sample_rate_0_5_none(mock_uuid, inputs, outputs, timestamps, sort, tags):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_predictions(
            application="foobar",
            inputs=inputs,
            outputs=outputs,
            feedback_keys=["A"],
            ignore_inputs=["B"],
            version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.5,
            as_batch=False,
            tags=tags,
        )
    log_store.log_batch.assert_not_called()


@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize(
    "timestamps", [pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME], None]
)
@pytest.mark.parametrize(
    "outputs",
    [
        [{"C": 300}, {"C": 303}],
        pd.DataFrame({"C": [300, 303]}),
    ],
)
@pytest.mark.parametrize(
    "inputs",
    [
        [{"A": 100, "B": 200}, {"A": 101, "B": 202}],
        pd.DataFrame({"A": [100, 101], "B": [200, 202]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_log_predictions_sample_rate_0(mock_uuid, inputs, outputs, timestamps, sort, tags):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.1):
        mock_uuid.return_value = "ABCD1234"
        gantry.log_predictions(
            application="foobar",
            inputs=inputs,
            outputs=outputs,
            feedback_keys=["A"],
            ignore_inputs=["B"],
            version=10,
            timestamps=timestamps,
            sort_on_timestamp=sort,
            sample_rate=0.0,
            as_batch=False,
            tags=tags,
        )
    log_store.log_batch.assert_not_called()


def test_log_predictions_mismatch_size_inputs_outputs(cli_obj):
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}, {"C": 303}],
        feedback_keys=["A"],
        ignore_inputs=["B"],
        version=10,
        as_batch=False,
    )

    assert cli_obj.log_store.logs == []


def test_log_predictions_mismatch_size_timestamp(cli_obj):
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["B"],
        version=10,
        timestamps=[SOME_TIME, ANOTHER_TIME],
        as_batch=False,
    )

    assert cli_obj.log_store.logs == []


def test_log_predictions_mismatch_size_feedback_ids(cli_obj):
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_ids=[{"foo": "bar"}, {"bar": "baz"}],
        ignore_inputs=["B"],
        version=10,
        as_batch=False,
    )

    assert cli_obj.log_store.logs == []


def test_log_predictions_mismatch_size_version(cli_obj):
    cli_obj.log_predictions(
        application="foobar",
        inputs=[{"A": 100, "B": 200}],
        outputs=[{"C": 300}],
        feedback_keys=["A"],
        ignore_inputs=["B"],
        version=[1, 2, 3],
        as_batch=False,
    )

    assert cli_obj.log_store.logs == []


def test_batch_id_population_without_batch(initialized_gantry, tmp_path):
    inputs = [{"input1": i, "input2": 2} for i in range(20)]
    outputs = [i % 2 for i in range(20)]

    initialized_gantry.log_predictions(
        "simple_test", inputs, outputs, sort_on_timestamp=True, as_batch=False
    )

    log_path = tmp_path / "predictions" / "simple_test.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_ev = json.loads(line)
            assert i < 20  # Max of 20 events
            assert logged_ev["batch_id"] is None


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("version", ["0", 0, None])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize("method", ["log_predictions", "log_records_prediction", "log_records"])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
@mock.patch("gantry.logger.client._batch_success_msg")
def test_log_predictions_as_batch(
    mock_batch_success, mock_uuid, version, sort, method, init_gantry_with_api
):
    if method == "log_records":
        size = 40
    else:
        size = 20
    mock_uuid.side_effect = [str(x) for x in range(size + 1)]

    inputs = [{"input1": i, "input2": 2} for i in range(20)]
    outputs = [i % 2 for i in range(20)]
    feedbacks = [{"val": i} for i in range(20)]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/ingest/file-preupload?filename=simple_test_{size}&num_parts=1&filetype=EVENTS",  # noqa
            json={
                "response": "ok",
                "upload_urls": ["http://test_url"],
                "upload_id": "10",
                "key": "x1234",
            },
        )
        resp.add(
            resp.PUT,
            "http://test_url",
            headers={"Content-Type": "application/json", "ETag": "working_etag"},
        )
        resp.add(
            resp.POST,
            url=f"{HOST}/api/v1/ingest/file",
            json={"response": "ok", "batch_id": "hello_123"},
            headers={"Content-Type": "application/json"},
        )

        if method == "log_predictions":
            rv = init_gantry_with_api.log_predictions(
                "simple_test", inputs, outputs, version, sort_on_timestamp=sort, as_batch=True
            )

            assert rv == "hello_123"
        elif method == "log_records_prediction":
            rv = init_gantry_with_api.log_records(
                "simple_test",
                version,
                inputs=inputs,
                outputs=outputs,
                sort_on_timestamp=sort,
                as_batch=True,
            )

            assert rv == "hello_123"
        elif method == "log_records":
            rv = init_gantry_with_api.log_records(
                "simple_test",
                version,
                inputs=inputs,
                outputs=outputs,
                feedbacks=feedbacks,
                sort_on_timestamp=sort,
                as_batch=True,
            )

            assert rv == "hello_123"
        mock_batch_success.assert_called_once_with(
            "hello_123", "simple_test", init_gantry_with_api.get_client().log_store
        )


@pytest.mark.parametrize("version", ["0", 0, None])
@pytest.mark.parametrize("sort", [True, False])
@pytest.mark.parametrize("method", ["log_feedback", "log_records_feedback"])
@pytest.mark.parametrize("feedback_id", [{"value": "foo"}, "foo"])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
@mock.patch("gantry.logger.client._batch_success_msg")
def test_log_feedback_as_batch(
    mock_batch_success, mock_uuid, version, sort, method, init_gantry_with_api, feedback_id
):
    size = 2
    mock_uuid.side_effect = [str(x) for x in range(size + 1)]
    feedbacks = [{"value": "foo"} for _ in range(size)]
    feedback_ids = [feedback_id for _ in range(size)]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/ingest/file-preupload?filename=simple_test_{size}&num_parts=1&filetype=EVENTS",  # noqa
            json={
                "response": "ok",
                "upload_urls": ["http://test_url"],
                "upload_id": "10",
                "key": "x1234",
            },
        )
        resp.add(
            resp.PUT,
            "http://test_url",
            headers={"Content-Type": "application/json", "ETag": "working_etag"},
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/ingest/file",
            json={"response": "ok", "batch_id": "hello_123"},
            headers={"Content-Type": "application/json"},
        )

        if method == "log_feedback":
            rv = init_gantry_with_api.log_feedback(
                application="simple_test",
                feedback_ids=feedback_ids,
                feedbacks=feedbacks,
                feedback_version=version,
                sort_on_timestamp=sort,
                as_batch=True,
            )

            assert rv == "hello_123"
        elif method == "log_records_feedback":
            rv = init_gantry_with_api.log_records(
                "simple_test",
                version,
                feedback_ids=feedback_ids,
                feedbacks=feedbacks,
                sort_on_timestamp=sort,
                as_batch=True,
            )

            assert rv == "hello_123"
        mock_batch_success.assert_called_once_with(
            "hello_123", "simple_test", init_gantry_with_api.get_client().log_store
        )


@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_batch_completion_fail(mock_uuid, init_gantry_with_api):
    size = 20
    inputs = [{"input1": i, "input2": 2} for i in range(size)]
    outputs = [i % 2 for i in range(size)]

    mock_uuid.side_effect = [str(x) for x in range(size + 1)]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/ingest/file-preupload?filename=simple_test_{size}&num_parts=1&filetype=EVENTS",  # noqa
            json={
                "response": "ok",
                "upload_urls": ["http://test_url"],
                "upload_id": "10",
                "key": "x1234",
            },
        )
        resp.add(
            resp.PUT,
            "http://test_url",
            headers={"Content-Type": "application/json", "ETag": "working_etag"},
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/ingest/file",
            json={"response": "not-ok"},
            headers={"Content-Type": "application/json"},
        )

        with pytest.raises(GantryBatchCreationException):
            init_gantry_with_api.log_predictions(
                "simple_test", inputs, outputs, sort_on_timestamp=True, as_batch=True
            )


@mock.patch("gantry.logger.event_builder.uuid.uuid4")
def test_batch_preupload_fail(mock_uuid, init_gantry_with_api):
    size = 20
    inputs = [{"input1": i, "input2": 2} for i in range(size)]
    outputs = [i % 2 for i in range(size)]
    mock_uuid.side_effect = [str(x) for x in range(size + 1)]

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/ingest/file-preupload?filename=simple_test_{size}&num_parts=1&filetype=EVENTS",  # noqa
            json={"response": "broken-return"},
        )

        with pytest.raises(GantryBatchCreationException):
            init_gantry_with_api.log_predictions(
                "simple_test", inputs, outputs, sort_on_timestamp=True, as_batch=True
            )


@pytest.mark.parametrize(
    ["inputs", "outputs", "feedbacks"],
    [
        ([{"A": 100}], None, [{"C": 100}]),
        (None, [{"A": 100}], [{"C": 100}]),
        (None, None, [{"C": 100}]),
        ([{"A": 100}], [{"C": 100}], None),
        ([{"A": 100}], None, None),
    ],
)
def test_log_predictions_and_feedbacks_no_preds_or_no_feedbacks(
    inputs, outputs, feedbacks, cli_obj, log_store
):
    cli_obj._log_predictions_and_feedbacks(
        application="foobar",
        version="2.0",
        inputs=inputs,
        outputs=outputs,
        feedback_keys=["A"],
        feedbacks=feedbacks,
    )

    assert log_store.logs == []


@pytest.mark.parametrize(
    ["inputs", "outputs", "feedbacks"],
    [
        ([{"A": 100}, {"A": 200}], [{"B": 100}], [{"B": 102}]),
        ([{"A": 100}, {"A": 200}], [{"B": 100}, {"B": 101}], [{"B": 102}]),
        ([{"A": 100}, {"A": 200}], [{"B": 100}, {"B": 101}], [{"B": 102}, {"B": 103}, {"B": 104}]),
        ([{"A": 100}], [{"B": 100}, {"B": 101}], [{"B": 102}, {"B": 103}, {"B": 104}]),
    ],
)
def test_log_predictions_and_feedbacks_size_mismatch(
    inputs, outputs, feedbacks, cli_obj, log_store
):
    cli_obj._log_predictions_and_feedbacks(
        application="foobar",
        version="2.0",
        inputs=inputs,
        outputs=outputs,
        feedback_keys=["A"],
        feedbacks=feedbacks,
    )

    assert log_store.logs == []


@pytest.mark.parametrize(
    "timestamps",
    [
        [],
        [SOME_TIME, ANOTHER_TIME],
        pd.DatetimeIndex([SOME_TIME, ANOTHER_TIME]),
    ],
)
def test_log_predictions_and_feedbacks_size_mismatch_timestamps(timestamps, cli_obj, log_store):
    cli_obj._log_predictions_and_feedbacks(
        application="foobar",
        version="2.0",
        inputs=[{"A": 100}],
        outputs=[{"B": 200}],
        feedback_keys=["A"],
        feedbacks=[{"A": 100}],
        timestamps=timestamps,
    )

    assert log_store.logs == []


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize(
    "timestamps", [None, pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME]]
)
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize("version", ["0", 0, None])
@pytest.mark.parametrize("as_batch", [False, True])
@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
    ],
)
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
@mock.patch("gantry.logger.client.Gantry._upload_data_as_batch")
def test_log_predictions_and_feedbacks_df(
    mock_upload_data_as_batch,
    mock_uuid,
    test_inputs,
    test_outputs,
    test_feedbacks,
    as_batch,
    version,
    sort,
    timestamps,
    tags,
    cli_obj,
    log_store,
):
    mock_upload_data_as_batch.return_value = None
    mock_uuid.side_effect = ["ABCD1", "ABCD2", "ABCD3", "ABCD4"]
    cli_obj._log_predictions_and_feedbacks(
        application="foobar",
        version=version,
        inputs=test_inputs,
        outputs=test_outputs,
        feedback_keys=["A"],
        feedbacks=test_feedbacks,
        timestamps=timestamps,
        sort_on_timestamp=sort,
        as_batch=as_batch,
        tags=tags,
    )
    expected_tags = tags if tags else {"env": "test"}

    event_1 = {
        "event_id": "ABCD2" if not (sort and timestamps is not None) else "ABCD4",
        "log_timestamp": CURRENT_TIME_STR,
        "timestamp": ANOTHER_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        "metadata": {
            "environment": tags["env"] if tags else "test",
            "func_name": "foobar",
            "version": version,
            "feedback_keys": ["A"],
            "ignore_inputs": None,
            "provided_feedback_id": None,
            "feedback_version": version,
        },
        "inputs": {"A": 100},
        "outputs": {"B": 300},
        "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
        "tags": expected_tags,
        "batch_id": None,
        "feedback_id_inputs": {"A": 100},
        "feedback": {"A": 200},
    }

    event_2 = {
        "event_id": "ABCD4" if not (sort and timestamps is not None) else "ABCD2",
        "log_timestamp": CURRENT_TIME_STR,
        "timestamp": SOME_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        "metadata": {
            "environment": tags["env"] if tags else "test",
            "func_name": "foobar",
            "version": version,
            "feedback_keys": ["A"],
            "ignore_inputs": None,
            "provided_feedback_id": None,
            "feedback_version": version,
        },
        "inputs": {"A": 101},
        "outputs": {"B": 301},
        "feedback_id": "9e329293e022d6cdaafdec49b5f4fedc",
        "tags": expected_tags,
        "batch_id": None,
        "feedback_id_inputs": {"A": 101},
        "feedback": {"A": 201},
    }

    if as_batch:
        mock_upload_data_as_batch.assert_called_once_with(
            "foobar",
            tags["env"] if tags else "test",
            version,
            [event_1, event_2] if not (sort and timestamps is not None) else [event_2, event_1],
            BatchType.RECORD,
        )
    else:
        assert log_store.logs == [
            {
                "events": [event_1, event_2]
                if not (sort and timestamps is not None)
                else [event_2, event_1],
                "application": "foobar",
            },
        ]
        mock_upload_data_as_batch.assert_not_called()


@freeze_time(CURRENT_TIME)
@pytest.mark.parametrize("tags", [{"env": "overwrite_env"}, {}, None])
@pytest.mark.parametrize(
    "timestamps", [None, pd.DatetimeIndex([ANOTHER_TIME, SOME_TIME]), [ANOTHER_TIME, SOME_TIME]]
)
@pytest.mark.parametrize("sort", [False, True])
@pytest.mark.parametrize("version", ["0", 0, None])
@mock.patch("gantry.logger.event_builder.uuid.uuid4")
@mock.patch("gantry.logger.client.Gantry._handle_upload")
def test__upload_data_as_batch(
    mock_handle_upload,
    mock_uuid,
    version,
    sort,
    timestamps,
    tags,
    cli_obj,
):
    mock_uuid.side_effect = ["1"]
    expected_tags = tags if tags else {"env": "test"}
    event_1 = {
        "event_id": "ABCD2" if not (sort and timestamps is not None) else "ABCD4",
        "log_timestamp": CURRENT_TIME_STR,
        "timestamp": ANOTHER_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        "metadata": {
            "environment": tags["env"] if tags else "test",
            "func_name": "foobar",
            "version": version,
            "feedback_keys": ["A"],
            "ignore_inputs": None,
            "provided_feedback_id": None,
            "feedback_version": version,
        },
        "inputs": {"A": 100},
        "outputs": {"B": 300},
        "feedback_id": "5dd14615efeb2d086e519ed35efd3f73",
        "tags": expected_tags,
        "batch_id": None,
        "feedback_id_inputs": {"A": 100},
        "feedback": {"A": 200},
    }
    event_2 = {
        "event_id": "ABCD4" if not (sort and timestamps is not None) else "ABCD2",
        "log_timestamp": CURRENT_TIME_STR,
        "timestamp": SOME_TIME_STR if timestamps is not None else CURRENT_TIME_STR,
        "metadata": {
            "environment": tags["env"] if tags else "test",
            "func_name": "foobar",
            "version": version,
            "feedback_keys": ["A"],
            "ignore_inputs": None,
            "provided_feedback_id": None,
            "feedback_version": version,
        },
        "inputs": {"A": 101},
        "outputs": {"B": 301},
        "feedback_id": "9e329293e022d6cdaafdec49b5f4fedc",
        "tags": expected_tags,
        "batch_id": None,
        "feedback_id_inputs": {"A": 101},
        "feedback": {"A": 201},
    }
    events = [event_1, event_2]

    cli_obj._upload_data_as_batch(
        "test_app", tags["env"] if tags else "test", version, events, BatchType.RECORD
    )

    mock_handle_upload.assert_called_once_with(
        mock.ANY,
        DataLink(
            application="test_app",
            version=str(version) if version is not None else version,
            file_type=UploadFileType.EVENTS,
            batch_type=BatchType.RECORD,
            log_timestamp=CURRENT_TIME.isoformat(),
            num_events=2,
            tags={"env": DataLinkElement(val=tags["env"] if tags else "test")},
        ),
        mock.ANY,
        "test_app_1",
    )


@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
def test_log_records_preds_and_feedback_series_coerced(
    mock_preds_and_feedback,
    cli_obj,
):
    expected_inputs = pd.Series([200, 201])
    expected_outputs = pd.Series([300, 301])
    expected_feedbacks = pd.Series([100, 101])
    application = "test_app"

    cli_obj.log_records(
        application=application,
        inputs=expected_inputs,
        outputs=expected_outputs,
        feedbacks=expected_feedbacks,
    )

    downstream_kwargs = mock_preds_and_feedback.call_args.kwargs
    actual_inputs = downstream_kwargs["inputs"]
    actual_outputs = downstream_kwargs["outputs"]
    actual_feedbacks = downstream_kwargs["feedbacks"]
    pd.testing.assert_frame_equal(actual_inputs, expected_inputs.to_frame())
    pd.testing.assert_frame_equal(actual_outputs, expected_outputs.to_frame())
    pd.testing.assert_frame_equal(actual_feedbacks, expected_feedbacks.to_frame())


@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
@mock.patch("gantry.logger.client.Gantry.log_predictions")
@mock.patch("gantry.logger.client.Gantry.log_feedback")
def test_log_records_preds_and_feedback(
    mock_feedback,
    mock_preds,
    mock_preds_and_feedback,
    version,
    test_inputs,
    test_outputs,
    test_feedbacks,
    cli_obj,
):
    rv = cli_obj.log_records(
        application="foobar",
        version=version,
        inputs=test_inputs,
        outputs=test_outputs,
        feedback_keys=["A"],
        feedbacks=test_feedbacks,
        timestamps=None,
        sort_on_timestamp=True,
    )
    assert rv is None

    mock_preds_and_feedback.assert_called_once()
    mock_preds.assert_not_called()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
def test_log_records_preds_and_feedback_0_5(version, test_inputs, test_outputs, test_feedbacks):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.4):
        gantry.log_records(
            application="foobar",
            version=version,
            inputs=test_inputs,
            outputs=test_outputs,
            feedback_keys=["A"],
            feedbacks=test_feedbacks,
            timestamps=None,
            sort_on_timestamp=True,
            sample_rate=0.5,
            as_batch=False,
        )

    log_store.log_batch.assert_called_once()


@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
def test_log_records_preds_and_feedback_0_5_none(
    version, test_inputs, test_outputs, test_feedbacks
):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        gantry.log_records(
            application="foobar",
            version=version,
            inputs=test_inputs,
            outputs=test_outputs,
            feedback_keys=["A"],
            feedbacks=test_feedbacks,
            timestamps=None,
            sort_on_timestamp=True,
            sample_rate=0.5,
            as_batch=False,
        )

    log_store.log_batch.assert_not_called()


@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
def test_log_records_preds_and_feedback_0_0(version, test_inputs, test_outputs, test_feedbacks):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    with mock.patch("random.random", return_value=0.6):
        gantry.log_records(
            application="foobar",
            version=version,
            inputs=test_inputs,
            outputs=test_outputs,
            feedback_keys=["A"],
            feedbacks=test_feedbacks,
            timestamps=None,
            sort_on_timestamp=True,
            sample_rate=0.0,
            as_batch=False,
        )

    log_store.log_batch.assert_not_called()


@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
@mock.patch("gantry.logger.client.Gantry.log_predictions")
@mock.patch("gantry.logger.client.Gantry.log_feedback")
def test_log_records_preds_only(
    mock_feedback, mock_preds, mock_preds_and_feedback, version, test_inputs, test_outputs, cli_obj
):
    rv = cli_obj.log_records(
        application="foobar",
        version=version,
        inputs=test_inputs,
        outputs=test_outputs,
        feedback_keys=["A"],
        feedbacks=None,
        timestamps=None,
        sort_on_timestamp=True,
    )
    assert rv is None

    mock_preds_and_feedback.assert_not_called()
    mock_preds.assert_called_once()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
    ],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
@mock.patch("gantry.logger.client.Gantry.log_predictions")
@mock.patch("gantry.logger.client.Gantry.log_feedback")
def test_log_records_feedback_only(
    mock_feedback, mock_preds, mock_preds_and_feedback, version, test_feedbacks, cli_obj
):
    rv = cli_obj.log_records(
        application="foobar",
        version=version,
        inputs=None,
        outputs=None,
        feedback_keys=["A"],
        feedbacks=test_feedbacks,
        timestamps=None,
        sort_on_timestamp=True,
    )
    assert rv is None

    mock_preds_and_feedback.assert_not_called()
    mock_preds.assert_not_called()
    mock_feedback.assert_called_once()


@pytest.mark.parametrize(
    "test_feedbacks",
    [None, pd.DataFrame(), [], pd.Series()],
)
@pytest.mark.parametrize(
    "test_outputs",
    [None, pd.DataFrame(), [], pd.Series()],
)
@pytest.mark.parametrize(
    "test_inputs",
    [None, pd.DataFrame(), [], pd.Series()],
)
@pytest.mark.parametrize("version", [None, 10, "1.2.3"])
@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
@mock.patch("gantry.logger.client.Gantry.log_predictions")
@mock.patch("gantry.logger.client.Gantry.log_feedback")
def test_log_records_no_data(
    mock_feedback,
    mock_preds,
    mock_preds_and_feedback,
    version,
    test_inputs,
    test_outputs,
    test_feedbacks,
    cli_obj,
):
    rv = cli_obj.log_records(
        application="foobar",
        version=version,
        inputs=test_inputs,
        outputs=test_outputs,
        feedback_keys=["A"],
        feedbacks=test_feedbacks,
        timestamps=None,
        sort_on_timestamp=True,
    )
    assert rv is None

    mock_preds_and_feedback.assert_not_called()
    mock_preds.assert_not_called()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_feedbacks",
    [None, pd.DataFrame(), [{"A": 3}, {"A": 3}], pd.Series()],
)
@mock.patch("gantry.logger.client.Gantry._log_predictions_and_feedbacks")
@mock.patch("gantry.logger.client.Gantry.log_predictions")
@mock.patch("gantry.logger.client.Gantry.log_feedback")
def test_log_records_incomplete_preds(
    mock_feedback,
    mock_preds,
    mock_preds_and_feedback,
    test_feedbacks,
    cli_obj,
):
    rv = cli_obj.log_records(
        application="foobar",
        version="2.0.1",
        inputs=[],
        outputs=[{"A": 1}, {"A": 2}],
        feedback_keys=["A"],
        feedbacks=test_feedbacks,
        timestamps=None,
        sort_on_timestamp=True,
    )
    assert rv is None

    mock_preds_and_feedback.assert_not_called()
    mock_preds.assert_not_called()
    mock_feedback.assert_not_called()


@mock.patch("gantry.logger.client.Gantry._log_prediction_and_feedback_event")
@mock.patch("gantry.logger.client.Gantry.log_prediction_event")
@mock.patch("gantry.logger.client.Gantry.log_feedback_event")
def test_log_record_pred_and_feedback(mock_feedback, mock_pred, mock_pred_and_feedback, cli_obj):
    cli_obj.log_record(
        application="foobar",
        version="2.0.1",
        inputs={"A": 100},
        outputs={"B": 300},
        feedback_keys=["A"],
        feedback={"A": 200},
        timestamp=None,
    )

    mock_pred_and_feedback.assert_called_once_with(
        application="foobar",
        version="2.0.1",
        inputs={"A": 100},
        outputs={"B": 300},
        feedback={"A": 200},
        feedback_keys=["A"],
        ignore_inputs=None,
        timestamp=None,
        sample_rate=1.0,
        tags=None,
    )
    mock_pred.assert_not_called()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_feedback",
    [None, {}],
)
@mock.patch("gantry.logger.client.Gantry._log_prediction_and_feedback_event")
@mock.patch("gantry.logger.client.Gantry.log_prediction_event")
@mock.patch("gantry.logger.client.Gantry.log_feedback_event")
def test_log_record_pred_only(
    mock_feedback, mock_pred, mock_pred_and_feedback, test_feedback, cli_obj
):
    cli_obj.log_record(
        application="foobar",
        version="2.0.1",
        inputs={"A": 100},
        outputs={"B": 300},
        feedback_keys=["A"],
        feedback=test_feedback,
        timestamp=None,
    )

    mock_pred_and_feedback.assert_not_called()
    mock_pred.assert_called_once_with(
        application="foobar",
        inputs={"A": 100},
        outputs={"B": 300},
        version="2.0.1",
        feedback_keys=["A"],
        ignore_inputs=None,
        feedback_id=None,
        timestamp=None,
        sample_rate=1.0,
        tags=None,
    )
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_output",
    [None, {}],
)
@pytest.mark.parametrize(
    "test_input",
    [None, {}],
)
@mock.patch("gantry.logger.client.Gantry._log_prediction_and_feedback_event")
@mock.patch("gantry.logger.client.Gantry.log_prediction_event")
@mock.patch("gantry.logger.client.Gantry.log_feedback_event")
def test_log_record_feedback_only(
    mock_feedback, mock_pred, mock_pred_and_feedback, test_input, test_output, cli_obj
):
    cli_obj.log_record(
        application="foobar",
        version="2.0.1",
        inputs=test_input,
        outputs=test_output,
        feedback_keys=["A"],
        feedback={"A": 200},
        timestamp=None,
    )

    mock_pred_and_feedback.assert_not_called()
    mock_pred.assert_not_called()
    mock_feedback.assert_called_once_with(
        application="foobar",
        feedback_id=None,
        feedback={"A": 200},
        feedback_version="2.0.1",
        timestamp=None,
        sample_rate=1.0,
        tags=None,
    )


@pytest.mark.parametrize(
    "test_input",
    [None, {}],
)
@mock.patch("gantry.logger.client.Gantry._log_prediction_and_feedback_event")
@mock.patch("gantry.logger.client.Gantry.log_prediction_event")
@mock.patch("gantry.logger.client.Gantry.log_feedback_event")
def test_log_record_no_data(mock_feedback, mock_pred, mock_pred_and_feedback, test_input, cli_obj):
    cli_obj.log_record(
        application="foobar",
        version="2.0.1",
        inputs=test_input,
        outputs={"B": 100},
        feedback_keys=["A"],
        feedback=None,
        timestamp=None,
    )

    mock_pred_and_feedback.assert_not_called()
    mock_pred.assert_not_called()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize("feedback", [None, {}, {"B": 200}])
@mock.patch("gantry.logger.client.Gantry._log_prediction_and_feedback_event")
@mock.patch("gantry.logger.client.Gantry.log_prediction_event")
@mock.patch("gantry.logger.client.Gantry.log_feedback_event")
def test_log_record_incomplete_preds(
    mock_feedback, mock_pred, mock_pred_and_feedback, feedback, cli_obj
):
    cli_obj.log_record(
        application="foobar",
        version="2.0.1",
        inputs={},
        outputs={"B": 100},
        feedback_keys=["A"],
        feedback=feedback,
        timestamp=None,
    )

    mock_pred_and_feedback.assert_not_called()
    mock_pred.assert_not_called()
    mock_feedback.assert_not_called()


@pytest.mark.parametrize(
    "test_inputs",
    [[{"A": 100}, {"A": 101}]],
)
@mock.patch("gantry.logger.utils.get_file_linecount")
def test_log_file(mock_line_count, test_inputs):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")
    mock_line_count.return_value = 5

    gantry.log_file(
        "foobar",
        "dir/csv_with_headers.csv",
        version="2.0.1",
    )

    log_store.log_batch.assert_not_called()


@pytest.mark.parametrize(
    "tags, expected_env",
    [
        [None, "test"],
        [{}, "test"],
        [{"other_tag": "random"}, "test"],
        [{"env": "overwrite_env"}, "overwrite_env"],
    ],
)
def test_get_env_and_context(tags, expected_env):
    log_store = mock.MagicMock()
    gantry = client.Gantry(log_store=log_store, environment="test")

    environment, context = gantry._get_env_and_context(tags)

    assert environment == expected_env
    assert context == {"environment": expected_env}



import uuid
from typing import Dict, List, Optional

from gantry.api_client import APIClient
from gantry.dataset.gantry_dataset import GantryDataset


class GantryDatasetClient:
    def __init__(
        self,
        api_client: APIClient,
        email: str,
        bucket_name: Optional[str] = None,
        aws_region: Optional[str] = None,
        workspace: str = ".",
    ):
        self._api_client = api_client
        self.email = email
        self.bucket_name = bucket_name
        self.aws_region = aws_region
        self.workspace = workspace

    def create_dataset(self, name: str, model_id: Optional[uuid.UUID] = None) -> GantryDataset:
        """
        Create dataset

        Args:
            name (str): dataset name
            model_id (Optional[uuid.UUID], optional): gantry model id which will be used to set
                dataset schema if provided.

        Returns:
            GantryDataset
        """
        data = {
            "name": name,
            "email": self.email,
        }
        if model_id:
            data["model_id"] = str(model_id)

        if self.bucket_name and self.aws_region:
            data.update(
                {
                    "bucket_name": self.bucket_name,
                    "aws_region": self.aws_region,
                }
            )
        elif not (not self.bucket_name and not self.aws_region):
            raise NotImplementedError(
                "Both bucket name and region need to be set if you want\
                 to use your own bucket."
            )

        res = self._api_client.request("POST", "/api/v1/datasets", json=data, raise_for_status=True)

        return GantryDataset(
            api_client=self._api_client,
            user_email=self.email,
            dataset_name=res["data"]["name"],
            dataset_id=res["data"]["id"],
            bucket_name=res["data"]["bucket_name"],
            aws_region=res["data"]["aws_region"],
            dataset_s3_prefix=f"{res['data']['organization_id']}/{res['data']['s3_prefix']}",
            workspace=self.workspace,
        )

    def get_dataset(self, name: str) -> GantryDataset:
        """
        Get dataset object

        Args:
            name (str): dataset name

        Returns:
            GantryDataset
        """
        res = self._api_client.request("GET", f"/api/v1/datasets/{name}", raise_for_status=True)

        return GantryDataset(
            api_client=self._api_client,
            user_email=self.email,
            dataset_name=res["data"]["name"],
            dataset_id=res["data"]["id"],
            bucket_name=res["data"]["bucket_name"],
            aws_region=res["data"]["aws_region"],
            dataset_s3_prefix=f"{res['data']['organization_id']}/{res['data']['s3_prefix']}",
            workspace=self.workspace,
        )

    def list_dataset_commits(self, name: str) -> List[Dict[str, str]]:
        """
        Show dataset commit history

        Args:
            name (str): dataset name

        Returns:
            list of commit info
        """
        dataset_id = self.get_dataset(name)._dataset_id
        res = self._api_client.request(
            "GET", f"/api/v1/datasets/{dataset_id}/commits", raise_for_status=True
        )

        return res["data"]

    def list_datasets(self) -> List[Dict[str, str]]:
        """
        List datasets

        Returns:
            List of dataset information
        """
        res = self._api_client.request("GET", "/api/v1/datasets", raise_for_status=True)

        return res["data"]



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from gantry.exceptions import GantryBatchCreationException
from gantry.logger.constants import BatchType, UploadFileType


@dataclass
class DataLinkElement:
    ref: Optional[int] = None
    val: Optional[Any] = None  # Exact value of the item.

    def __post_init__(self):
        if self.ref is None and self.val is None:
            raise GantryBatchCreationException("ref or val must be populated.")
        return self


@dataclass
class DataLink:
    file_type: UploadFileType
    batch_type: BatchType
    num_events: int
    application: str
    version: Optional[str]
    batch_id: Optional[str] = None
    log_timestamp: Optional[str] = None
    timestamp: Dict[str, DataLinkElement] = field(default_factory=dict)
    inputs: Dict[str, DataLinkElement] = field(default_factory=dict)
    outputs: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback: Dict[str, DataLinkElement] = field(default_factory=dict)
    tags: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback_id: Dict[str, DataLinkElement] = field(default_factory=dict)
    feedback_keys: List[str] = field(default_factory=list)



from enum import Enum


class BatchType(str, Enum):
    RECORD = "RECORD"
    FEEDBACK = "FEEDBACK"
    PREDICTION = "PREDICTION"


class UploadFileType(str, Enum):
    CSV_WITH_HEADERS = "CSV_WITH_HEADERS"
    EVENTS = "EVENTS"


class Delimiter(Enum):
    COMMA = ","


CHUNK_SIZE = 100 * 1024 * 1024  # File upload chunk size 100 MB



from gantry.logger.event_builder import _build_data_link
from gantry.logger.constants import BatchType, UploadFileType


def test_build_data_link(datadir):
    csv_file = datadir / "feedback.csv"
    data_link = _build_data_link(
        application="test_app",
        version=0,
        environment="dev",
        feedback_id=["feedback_id"],
        feedbacks=["test_float", "test_int"],
        filepath=str(csv_file),
    )

    assert data_link.application == "test_app"
    assert data_link.version == "0"
    assert data_link.file_type == UploadFileType.CSV_WITH_HEADERS
    assert data_link.batch_type == BatchType.FEEDBACK
    assert data_link.tags["env"].val == "dev"
    assert data_link.feedback_id["feedback_id"].ref == 0
    assert data_link.feedback["test_float"].ref == 1
    assert data_link.feedback["test_int"].ref == 2



import json
from io import StringIO
from pathlib import Path

import pytest
import responses
from responses import matchers

from gantry.api_client import APIClient
from gantry.dataset.constants import DELETED_FILES, MODIFIED_FILES, NEW_FILES
from gantry.dataset.gantry_dataset import DatasetFileInfo, GantryDataset
from gantry.exceptions import GantryException

from .conftest import (
    AWS_REGION,
    BUCKET_NAME,
    COMMIT_MSG,
    CONF_OBJ_KEY,
    CSV_OBJ_KEY,
    DATASET_ID,
    DATASET_NAME,
    HOST,
    IMG_OBJ_KEY,
    MANIFEST_OBJ_KEY,
    MANIFEST_VERSION_ID,
    README_OBJ_KEY,
    S3_PREFIX,
    USER_EMAIL,
)


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture(scope="function")
def gantry_dataset_obj(test_api_client, datadir):
    return GantryDataset(
        api_client=test_api_client,
        dataset_name=DATASET_NAME,
        user_email=USER_EMAIL,
        dataset_id=DATASET_ID,
        bucket_name=BUCKET_NAME,
        aws_region=AWS_REGION,
        dataset_s3_prefix=S3_PREFIX,
        workspace=datadir,
    )


def test_list_commits(gantry_dataset_obj, commit_history):
    """
    Test get dataset commits
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        commits = gantry_dataset_obj.list_commits()
        assert len(commits) == 2


def test_get_commit(gantry_dataset_obj, commit_history):
    """
    Test get commit information
    """
    expected_commit = commit_history[1]
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits/{expected_commit['id']}",
            json={
                "response": "ok",
                "data": expected_commit,
            },
            headers={"Content-Type": "application/json"},
        )
        commit = gantry_dataset_obj.get_commit(expected_commit["id"])
        assert commit == expected_commit


def test_get_latest_commit(gantry_dataset_obj, commit_history):
    """
    Test get latest commit
    """
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
            json={
                "response": "ok",
                "data": commit_history,
            },
            headers={"Content-Type": "application/json"},
        )
        latest_commit = gantry_dataset_obj.get_latest_commit()
        assert latest_commit["id"] == "8ee0f6d5-c84c-473a-931a-5148b2e704d7"
        assert latest_commit["is_latest_commit"]


def test_get_diff(gantry_dataset_obj):
    """
    Test get local edit diff
    """
    # overwrite the dataset name
    gantry_dataset_obj.dataset_name = "show_diff_dataset"

    diff = gantry_dataset_obj.get_diff()

    assert set(diff[NEW_FILES]) == {"artifacts/kitten.png", "tabular_manifests/feedback.csv"}
    assert diff[MODIFIED_FILES] == ["dataset_config.yaml"]
    assert diff[DELETED_FILES] == ["README.md"]


def setup_resp_for_create_commit(
    datadir, resp, create_commit_resp_json, create_commit_status_code=200
):
    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/presign/putobject",
        json={
            "response": "ok",
            "data": {
                MANIFEST_OBJ_KEY: f"{HOST}/fake_presigned_url",
                CONF_OBJ_KEY: f"{HOST}/fake_presigned_url",
                README_OBJ_KEY: f"{HOST}/fake_presigned_url",
                IMG_OBJ_KEY: f"{HOST}/fake_presigned_url",
                CSV_OBJ_KEY: f"{HOST}/fake_presigned_url",
            },
        },
        headers={"Content-Type": "application/json"},
    )

    resp.add(
        resp.PUT,
        f"{HOST}/fake_presigned_url",
        json={"response": "ok"},
        headers={"Content-Type": "application/json", "x-amz-version-id": MANIFEST_VERSION_ID},
    )

    with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
        old_commit = json.load(f)

    resp.add(
        resp.POST,
        f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
        json=create_commit_resp_json,
        headers={"Content-Type": "application/json"},
        status=create_commit_status_code,
        match=[
            matchers.json_params_matcher(
                {
                    "message": COMMIT_MSG,
                    "metadata_s3_file_version": MANIFEST_VERSION_ID,
                    "parent_commit_id": old_commit["id"],
                    "email": USER_EMAIL,
                }
            )
        ],
    )


def test_create_commit(datadir, gantry_dataset_obj, commit_history):
    """
    Test create commit succeed
    """
    with responses.RequestsMock() as resp:
        setup_resp_for_create_commit(datadir, resp, {"response": "ok", "data": commit_history[0]})

        gantry_dataset_obj.create_commit(COMMIT_MSG)

        diff = gantry_dataset_obj.get_diff()

        # verify no diff after a successful commit
        assert not diff[NEW_FILES]
        assert not diff[MODIFIED_FILES]
        assert not diff[DELETED_FILES]

        with open(datadir / f"{DATASET_NAME}/.dataset_metadata/HEAD") as f:
            assert json.load(f) == commit_history[0]  # verify HEAD has been updated


def test_create_commit_failure(datadir, gantry_dataset_obj):
    """
    Test create commit failure, in this case all upload will finish successfully but the create
    commit call will fail. This will happen during a race condition when another user committed a
    change and the local copy out of date.
    """
    with responses.RequestsMock() as resp:
        setup_resp_for_create_commit(
            datadir, resp, {"response": "error", "error": "Parent commit out of date!"}, 400
        )

        with pytest.raises(GantryException):
            gantry_dataset_obj.create_commit(COMMIT_MSG)

        diff = gantry_dataset_obj.get_diff()

        # since commit failed the local diff will be the same
        assert set(diff[NEW_FILES]) == {"artifacts/kitten.png", "tabular_manifests/feedback.csv"}
        assert diff[MODIFIED_FILES] == ["dataset_config.yaml"]
        assert diff[DELETED_FILES] == ["README.md"]


def test_create_commit_nochange(datadir, gantry_dataset_obj):
    """
    Test create commit without any local change
    """
    gantry_dataset_obj.dataset_name = "unittest_dataset"
    with open(datadir / "unittest_dataset/.dataset_metadata/HEAD") as f:
        old_commit = json.load(f)

    new_commit = gantry_dataset_obj.create_commit("new commit")
    assert old_commit == new_commit


def create_mock_manifest(commit_history):
    gantry_manifest = StringIO()
    config_file_info = {
        "file_name": "dataset_config.yaml",
        "url": f"s3://test-bucket/{CONF_OBJ_KEY}",
        "sha256": "unmatched_sha256",
        "version_id": "random_vid_for_config",
    }

    gantry_manifest.write(f"{json.dumps(config_file_info)}\n")

    readme_file_info = {
        "file_name": "README.md",
        "url": f"s3://test-bucket/{README_OBJ_KEY}",
        "sha256": "unmatched_sha256",
        "version_id": "random_vid_for_readme",
    }

    gantry_manifest.write(f"{json.dumps(readme_file_info)}\n")
    gantry_manifest.seek(0)
    commit_json = commit_history[1]

    return gantry_manifest, commit_json


@pytest.mark.parametrize("commit_id", [None, "eb2e7242-3340-4edf-8366-90d4fce897ce"])
def test_sync_dataset(commit_id, datadir, gantry_dataset_obj, commit_history):
    """
    Test sync local data set
    1. To latest commit
    2. To a specific commit
    """
    manifest_file, commit_json = create_mock_manifest(commit_history)

    gantry_dataset_obj.dataset_name = "show_diff_dataset"

    with responses.RequestsMock() as resp:
        if commit_id is not None:
            resp.add(
                resp.GET,
                f"{HOST}/api/v1/datasets/{DATASET_ID}/commits/{commit_id}",
                json={
                    "response": "ok",
                    "data": commit_json,
                },
                headers={"Content-Type": "application/json"},
            )
        else:
            resp.add(
                resp.GET,
                f"{HOST}/api/v1/datasets/{DATASET_ID}/commits",
                json={
                    "response": "ok",
                    "data": [commit_json],
                },
                headers={"Content-Type": "application/json"},
            )

        resp.add(
            resp.POST,
            f"{HOST}/api/v1/datasets/{DATASET_ID}/presign/getobject",
            json={
                "response": "ok",
                "data": {
                    MANIFEST_OBJ_KEY: f"{HOST}/fake_manifest",
                    CONF_OBJ_KEY: f"{HOST}/fake_presigned_url",
                    README_OBJ_KEY: f"{HOST}/fake_presigned_url",
                },
            },
            headers={"Content-Type": "application/json"},
        )

        req_kwargs = {
            "stream": True,
        }
        resp.add(
            resp.GET,
            f"{HOST}/fake_presigned_url",
            body="fake_response",
            match=[matchers.request_kwargs_matcher(req_kwargs)],
        )

        resp.add(
            resp.GET,
            f"{HOST}/fake_manifest",
            body=manifest_file.read(),
            match=[matchers.request_kwargs_matcher(req_kwargs)],
        )
        commit = gantry_dataset_obj.sync(commit_id)
        assert commit == commit_json
        with open(datadir / "show_diff_dataset/dataset_config.yaml", "r") as f:
            assert f.read() == "fake_response"

        with open(datadir / "show_diff_dataset/README.md", "r") as f:
            assert f.read() == "fake_response"

        with open(datadir / "show_diff_dataset/.dataset_metadata/HEAD") as f:
            assert json.load(f) == commit_json

        assert not any(Path(datadir / "show_diff_dataset/tabular_manifests").iterdir())
        assert not any(Path(datadir / "show_diff_dataset/artifacts").iterdir())


def test_to_jsonl():
    file_info = DatasetFileInfo(
        file_name="test_file",
        url="s3://test-bucket/key",
        version_id="version_id",
        sha256="mock_sha256",
    )
    assert (
        file_info.to_jsonl()
        == '{"file_name": "test_file", "url": "s3://test-bucket/key", "version_id": "version_id", \
"sha256": "mock_sha256"}\n'
    )


@pytest.mark.parametrize(
    ["file_name", "url", "version_id", "sha256"],
    [
        (None, "s3://test-bucket/key", "version_id", "mock_sha256"),
        ("test_file", None, "version_id", "mock_sha256"),
        ("test_file", "s3://test-bucket/key", None, "mock_sha256"),
        ("test_file", "s3://test-bucket/key", "version_id", None),
    ],
)
def test_to_jsonl_value_error(file_name, url, version_id, sha256):
    file_info = DatasetFileInfo(
        file_name=file_name,
        url=url,
        version_id=version_id,
        sha256=sha256,
    )
    with pytest.raises(ValueError):
        file_info.to_jsonl()



import datetime
import logging
import random
import uuid
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd

from gantry.exceptions import GantryLoggingException
from gantry.logger.constants import BatchType, Delimiter, UploadFileType
from gantry.logger.types import DataLink, DataLinkElement
from gantry.logger.utils import _check_sample_rate, get_file_linecount
from gantry.utils import check_event_time_in_future, compute_feedback_id
from gantry.validators import validate_logged_field_datatype

logger = logging.getLogger(__name__)


def _build_prediction_and_feedback_events(
    application: str,
    context: Dict[str, str],
    environment: str,
    timestamp_idx: Iterable,
    inputs: Union[List[dict], pd.DataFrame],
    outputs: Union[List[dict], List[Any], pd.DataFrame],
    feedbacks: Union[List[dict], pd.DataFrame],
    version: Optional[Union[str, int]] = None,
    feedback_keys: Optional[List[str]] = None,
    ignore_inputs: Optional[List[str]] = None,
    sample_rate: float = 1.0,
    tags: Optional[Dict[str, str]] = None,
):
    """Build prediction and feedback events for log_records"""
    _check_sample_rate(sample_rate)

    events = []
    for idx, timestamp in timestamp_idx:
        if random.random() > sample_rate:
            continue

        feedback_id = {k: inputs[idx][k] for k in (feedback_keys or [])}

        event = {}

        event.update(
            _build_prediction_event(
                context,
                environment,
                inputs[idx],
                outputs[idx],
                application,
                version,
                feedback_keys,
                ignore_inputs,
                feedback_id=None,
                custom_timestamp=timestamp,
                tags=tags,
            )
        )

        feedback_event = _build_feedback_event(
            context,
            application,
            feedback_id,
            feedbacks[idx],
            version,  # type: ignore
            timestamp,
        )

        # Build an event with prediction data
        # and feedback data, now that backend
        # supports it.
        event["metadata"].update(feedback_event.pop("metadata"))
        event.update(feedback_event)

        events.append(event)

    return events


def _create_timestamp_idx(
    sort_on_timestamp: bool, timestamps: Optional[Iterable[Any]], record_count: int
) -> Iterable:
    """
    This function is used to take an index of timestamps, possibly None, and a length,
    and turn it into a mapping from the current timestamp to the original index of the
    corresponding value. The goal is to send all data to Gantry in timestamp order to
    minimize summarization window computing overhead by packing events of a single
    window into a single task.
    """
    if timestamps is not None:
        timestamp_idx: Iterable[Tuple[int, Any]] = enumerate(timestamps)
        if sort_on_timestamp:
            timestamp_idx = sorted(timestamp_idx, key=lambda el: el[1])
    else:
        timestamp = datetime.datetime.utcnow()
        timestamp_idx = ((i, timestamp) for i in range(record_count))

    return timestamp_idx


def _enrich_events_with_batch_id(events, batch_id):
    for event in events:
        event["batch_id"] = batch_id


def _build_feedback_event(
    context: dict,
    application: str,
    feedback_id: Union[str, dict],
    feedback: dict,
    feedback_version: Optional[Union[str, int]] = None,
    timestamp: Optional[datetime.datetime] = None,
    batch_id: Optional[str] = None,
) -> dict:
    """
    Create a feedback event record for logging.
    """
    metadata: Dict[str, Any] = {}
    metadata.update(context)
    metadata.update({"func_name": application, "feedback_version": feedback_version})
    if isinstance(feedback_id, str):
        computed_feedback_id = feedback_id
    else:
        computed_feedback_id = compute_feedback_id(feedback_id, list(feedback_id.keys()))

    log_time = datetime.datetime.utcnow()
    event_time = timestamp if timestamp else log_time
    if check_event_time_in_future(event_time):
        raise GantryLoggingException(
            "Cannot log events from the future. "
            f"Event timestep is {event_time}, but current time is {log_time}. "
            "Please check your event times and re-submit all."
        )
    log_time_str, event_time_str = log_time.isoformat(), event_time.isoformat()

    validate_logged_field_datatype(feedback, paths=["feedback"])
    validate_logged_field_datatype(feedback_id, paths=["feedback_id"])
    validate_logged_field_datatype(event_time_str, paths=["timestamp"])

    return {
        "event_id": uuid.uuid4(),
        "timestamp": event_time_str,
        "log_timestamp": log_time_str,
        "metadata": metadata,
        "feedback_id_inputs": feedback_id,
        "feedback": feedback,
        "feedback_id": computed_feedback_id,
        "batch_id": batch_id,
    }


def _build_prediction_events(
    application: str,
    inputs: Union[List[dict], pd.DataFrame],
    outputs: Union[List[dict], List[Any], pd.DataFrame],
    timestamp_idx: Iterable,
    context: Dict[str, str],
    environment: str,
    version: Optional[Union[str, int]],
    feedback_keys: Optional[List[str]] = None,
    ignore_inputs: Optional[List[str]] = None,
    feedback_ids: Optional[Union[List[str], List[dict]]] = None,
    sample_rate: float = 1.0,
    batch_id: Optional[str] = None,
    tags: Optional[Dict[str, str]] = None,
):
    _check_sample_rate(sample_rate)

    events = []
    for idx, timestamp in timestamp_idx:
        if random.random() > sample_rate:
            continue

        feedback_id = feedback_ids[idx] if feedback_ids else None

        events.append(
            _build_prediction_event(
                context,
                environment,
                inputs[idx],
                outputs[idx],
                application,
                version,
                feedback_keys,
                ignore_inputs,
                batch_id=batch_id,
                feedback_id=feedback_id,
                custom_timestamp=timestamp,
                tags=tags,
            )
        )

    return events


def _build_prediction_event(
    context: dict,
    env: str,
    inputs: Dict,
    outputs: Any,
    application: str,
    version: Optional[Union[int, str]],
    feedback_keys: Optional[List[str]],
    ignore_inputs: Optional[List[str]],
    batch_id: Optional[str] = None,
    feedback_id: Optional[Union[str, dict]] = None,
    tags: Optional[Dict[str, str]] = None,
    custom_timestamp: Optional[datetime.datetime] = None,
):
    metadata = {}
    metadata.update(context)
    metadata.update(
        {
            "func_name": application,
            "version": version,
            "feedback_keys": feedback_keys,
            "ignore_inputs": ignore_inputs,
            "provided_feedback_id": feedback_id,
        }
    )
    inputs = dict(inputs)  # make a copy so we can pop keys safely

    if feedback_id:
        if feedback_keys:
            raise GantryLoggingException(
                "Cannot specify feedback_id and feedback_keys at same time."
            )

        if isinstance(feedback_id, str):
            # Set the feedback_id directly from the given string.
            computed_feedback_id = feedback_id
        else:
            # Compute feedback_id just from provided feedback_id dictionary
            computed_feedback_id = compute_feedback_id(feedback_id, list(feedback_id.keys()))
    else:
        computed_feedback_id = compute_feedback_id(inputs, feedback_keys)

    if ignore_inputs:
        for ignore in ignore_inputs:
            inputs.pop(ignore, None)

    log_time = datetime.datetime.utcnow()
    event_time = custom_timestamp if custom_timestamp else log_time
    if check_event_time_in_future(event_time):
        raise GantryLoggingException(
            "Cannot log events from the future. "
            f"Event timestep is {event_time}, but current time is {log_time}. "
            "Please check your event times and re-submit all."
        )
    log_time_str, event_time_str = log_time.isoformat(), event_time.isoformat()

    processed_tags = {"env": env}
    if tags:
        processed_tags.update(tags)

    validate_logged_field_datatype(inputs, paths=["inputs"])
    validate_logged_field_datatype(outputs, paths=["outputs"])

    return {
        "event_id": uuid.uuid4(),
        "log_timestamp": log_time_str,
        "timestamp": event_time_str,
        "metadata": metadata,
        "inputs": inputs,
        "outputs": outputs,
        "feedback_id": computed_feedback_id,
        "tags": processed_tags,
        "batch_id": batch_id,
    }


def _build_data_link(  # noqa: C901
    application: str,
    environment: Optional[str] = None,
    version: Optional[Union[int, str]] = None,
    timestamp: Optional[str] = None,
    inputs: List[str] = [],
    outputs: List[str] = [],
    feedbacks: List[str] = [],
    tags: List[str] = [],
    feedback_id: List[str] = [],
    feedback_keys: Optional[List[str]] = None,
    filepath: Optional[str] = None,
    events: Optional[List] = None,
    batch_type: BatchType = BatchType.RECORD,
) -> DataLink:
    if filepath:
        # Get info for tracking batch during processing.
        num_events = get_file_linecount(filepath) - 1  # Subtract the header line.
        file_type = UploadFileType.CSV_WITH_HEADERS
    elif events:
        num_events = len(events)
        file_type = UploadFileType.EVENTS
    else:
        raise GantryLoggingException("Invalid data params. Need filepath or events list")
    logger.info(f"Processing {num_events} lines.")

    if version is not None:
        version = str(version)

    # Read the headers of the file, infer and build linking datastruct.
    data_link = DataLink(
        file_type=file_type,
        batch_type=batch_type,
        num_events=num_events,
        application=application,
        version=version,
        log_timestamp=datetime.datetime.utcnow().isoformat(),
        tags={"env": DataLinkElement(val=environment)},
    )

    if filepath:
        with open(filepath, "r") as f_in:
            headers_line = f_in.readline().split(Delimiter.COMMA.value)
            data_line = f_in.readline().split(Delimiter.COMMA.value)
            if len(headers_line) != len(data_line):
                raise GantryLoggingException("Header row length does not match data row length.")

            for col_index, raw_header in enumerate(headers_line):
                header = raw_header.strip()
                if header == timestamp or (not timestamp and header.startswith("timestamp")):
                    data_link.timestamp[header] = DataLinkElement(ref=col_index)
                elif header in feedback_id or header == "feedback_id":
                    data_link.feedback_id[header] = DataLinkElement(ref=col_index)
                elif header in outputs or header.startswith("output"):
                    data_link.outputs[header] = DataLinkElement(ref=col_index)
                elif header in feedbacks or header.startswith("feedback"):
                    data_link.feedback[header] = DataLinkElement(ref=col_index)
                elif header in tags or header.startswith("tags"):
                    data_link.tags[header] = DataLinkElement(ref=col_index)
                elif len(inputs) == 0 or header in inputs or header.startswith("input"):
                    data_link.inputs[header] = DataLinkElement(ref=col_index)

                if feedback_keys is not None and header in feedback_keys:
                    data_link.feedback_keys.append(header)

        # Overwrite the batch type for feedback events
        # TODO:// we need to have a better logic to handle different edge cases:
        # 1. provided input without output
        # 2. provided output without input
        # 3. customer specified batch type doesn't match the column
        # 4. conflicts between column params and column name. For example what if customer provide
        #    feedback = [output_feedback] <- output_feedback will be treated as output now.
        if (
            not data_link.inputs
            and not data_link.outputs
            and data_link.feedback
            and data_link.feedback_id
        ):
            data_link.batch_type = BatchType.FEEDBACK

    if data_link.tags["env"].ref is None and data_link.tags["env"].val is None:
        raise GantryLoggingException(
            "Please initialize gantry logger with 'environment' or specify an 'env' tag."
        )

    return data_link



import datetime
import json
import logging
import uuid

import dateutil.parser
import pytest
from mock import MagicMock, patch

from gantry.exceptions import ClientNotInitialized
from gantry.logger.client import Gantry
from gantry.utils import compute_feedback_id


def test_global_client():
    log_store = MagicMock()
    gantry = Gantry(log_store=log_store, environment="test")

    @gantry.instrument("simple_test")
    def test(input1: int, input2: str) -> dict:
        return {"output1": 1, "output2": "2"}

    log_store.log.assert_not_called()

    fixed_uuid = uuid.uuid4()
    with patch("uuid.uuid4", return_value=fixed_uuid):
        test(1, "2")

    log_store.log.assert_called_once()

    call_args = log_store.log.call_args[0]

    assert call_args[0] == "simple_test"
    assert call_args[1]["event_id"] == fixed_uuid
    assert call_args[1]["inputs"] == {"input1": 1, "input2": "2"}
    assert call_args[1]["outputs"] == {"output1": 1, "output2": "2"}
    assert call_args[1]["metadata"]["func_name"] == "simple_test"
    assert call_args[1]["metadata"]["version"] is None


def test_global_client_not_init(gantry):
    with pytest.raises(ClientNotInitialized):

        @gantry.instrument("simple_test")
        def test(input1: int, input2: str) -> dict:
            return dict(input1=input1, input2=input2)


def test_client_sampling():
    log_store = MagicMock()
    gantry = Gantry(log_store=log_store, environment="test")

    @gantry.instrument("simple_test", sample_rate=0.3)
    def test(input1: int, input2: str) -> dict:
        return {"output1": 1, "output2": "2"}

    with patch("random.random", return_value=0.4):
        test(1, "2")

    log_store.log.assert_not_called()

    with patch("random.random", return_value=0.2):
        test(1, "2")

    log_store.log.assert_called_once()


def test_client_exception_continue(gantry, caplog):
    caplog.set_level(logging.ERROR, logger="gantry.client")

    log_store = MagicMock()
    log_store.log.side_effect = Exception("Bad thing happened.")

    gantry = Gantry(log_store=log_store, environment="test")

    test_ret_val = {"output1": 1, "output2": "2"}

    @gantry.instrument("simple_test")
    def test(input1: int, input2: str) -> dict:
        return test_ret_val

    ret_val = test(1, "2")

    assert ret_val == test_ret_val
    assert "Internal exception in Gantry client" in caplog.text


def test_client_log_feedback_event(tmp_path, initialized_gantry):
    initialized_gantry.log_feedback_event("test_name", {"test_id": 1}, {"label": True})
    log_path = tmp_path / "feedback" / "test_name.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_feedback = json.loads(line)
            assert i == 0  # There is only 1 event.
            assert logged_feedback["feedback"]["label"]


def test_client_log_feedback(tmp_path, initialized_gantry):
    ids = [{"test_id": i} for i in range(20)]
    feedbacks = [{"label": i} for i in range(20)]
    initialized_gantry.log_feedback("test_name", ids, feedbacks, as_batch=False)

    log_path = tmp_path / "feedback" / "test_name.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_feedback = json.loads(line)
            assert i < 20  # There should only be 20 events.
            assert i == logged_feedback["feedback"]["label"]


def test_client_log_prediction_event(tmp_path, initialized_gantry):
    inputs = {"input1": 1, "input2": 2}
    outputs = 0
    initialized_gantry.log_prediction_event("simple_test", inputs, outputs)

    log_path = tmp_path / "predictions" / "simple_test.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_ev = json.loads(line)
            assert i == 0  # There is only 1 event
            assert logged_ev["log_timestamp"] == logged_ev["timestamp"]


def test_client_log_prediction_custom_timestamp(tmp_path, initialized_gantry):
    inputs = {"input1": 1, "input2": 2}
    outputs = 0

    custom_timestamp = datetime.datetime(year=2020, month=1, day=1)
    initialized_gantry.log_prediction_event(
        "simple_test", inputs, outputs, timestamp=custom_timestamp
    )

    log_path = tmp_path / "predictions" / "simple_test.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_ev = json.loads(line)
            assert i == 0  # Only 1 event
            assert logged_ev["timestamp"] == custom_timestamp.isoformat()
            diff = datetime.datetime.utcnow() - dateutil.parser.isoparse(logged_ev["log_timestamp"])
            assert diff < datetime.timedelta(minutes=1)  # log timestamp is recent


def test_client_log_prediction_feedback_id(tmp_path, initialized_gantry, caplog):
    inputs = {"input1": 1, "input2": 2}
    outputs = 0

    initialized_gantry.log_prediction_event(
        "simple_test", inputs, outputs, feedback_id={"id": "123"}
    )

    log_path = tmp_path / "predictions" / "simple_test.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_ev = json.loads(line)
            assert i == 0  # Only 1 event
            assert logged_ev["feedback_id"] == compute_feedback_id({"id": "123"})

    initialized_gantry.log_prediction_event(
        "simple_test", inputs, outputs, feedback_id={"id": "123"}, feedback_keys=["id"]
    )
    assert "Cannot specify feedback_id and feedback_keys at same time." in caplog.text


def test_client_log_predictions(tmp_path, initialized_gantry):
    _test_client_log_predictions_helper(initialized_gantry, tmp_path, False)


def test_client_log_predictions_sort_inputs(tmp_path, initialized_gantry):
    _test_client_log_predictions_helper(initialized_gantry, tmp_path, True)


def _test_client_log_predictions_helper(gantry, temp_file_path, sort_inputs):
    inputs = [{"input1": i, "input2": 2} for i in range(20)]
    outputs = [i % 2 for i in range(20)]

    gantry.log_predictions(
        "simple_test", inputs, outputs, sort_on_timestamp=sort_inputs, as_batch=False
    )

    log_path = temp_file_path / "predictions" / "simple_test.log"
    assert log_path.exists()

    with log_path.open() as f:
        for i, line in enumerate(f):
            logged_ev = json.loads(line)
            assert i < 20  # Max of 20 events
            assert logged_ev["inputs"]["input1"] == i
            assert logged_ev["outputs"] == i % 2
            assert logged_ev["metadata"]["func_name"] == "simple_test"



import json
import logging
import os
import shutil
import uuid
from collections import defaultdict
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List, Optional

from gantry.api_client import APIClient
from gantry.dataset import constants
from gantry.exceptions import GantryRequestException
from gantry.utils import (
    download_file_from_url,
    get_files_checksum,
    parse_s3_path,
    upload_file_to_url,
)

logger = logging.getLogger(__name__)

BATCH_SIZE = 5


@dataclass
class DatasetFileInfo:
    file_name: str
    url: str
    version_id: Optional[str] = None
    sha256: Optional[str] = None

    def to_jsonl(self):
        if not self.file_name or not self.url or not self.version_id or not self.sha256:
            raise ValueError(
                f"Failed to create file info jsonl: {json.dumps(asdict(self))}. \
                    Incomplete dataset file info line! "
            )

        return f"{json.dumps(asdict(self))}\n"


class GantryDataset:
    def __init__(
        self,
        api_client: APIClient,
        dataset_name: str,
        user_email: str,
        dataset_id: uuid.UUID,
        bucket_name: str,
        aws_region: str,
        dataset_s3_prefix: str,
        workspace: str,
    ):
        self.dataset_name = dataset_name
        self.user_email = user_email
        self.workspace = workspace
        self._api_client = api_client
        self._dataset_id = dataset_id
        self._bucket_name = bucket_name
        self._dataset_s3_prefix = dataset_s3_prefix
        self._aws_region = aws_region

    def list_commits(self) -> List[Dict[str, str]]:
        """
        Get dataset commits history

        Returns:
            List[Dict[str, str]]: dataset commits sorted from latest to earlist.
        """
        response = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits", raise_for_status=True
        )
        return response["data"]

    def get_diff(self) -> Dict[str, List[str]]:
        """
        Get local changes which hasn't been committed
        Return:
            {
                "new_files": List[str],
                "modified_files": List[str],
                "deleted_files": List[str],
            }
        """
        diff = self._get_diff()
        return {
            constants.NEW_FILES: [f.file_name for f in diff[constants.NEW_FILES]],
            constants.MODIFIED_FILES: [f.file_name for f in diff[constants.MODIFIED_FILES]],
            constants.DELETED_FILES: [f.file_name for f in diff[constants.DELETED_FILES]],
        }

    def create_commit(self, message: str) -> Dict[str, str]:
        """
        This method will compute the difference between the current commit and the local directory,
        upload new files and modified files to dataset repo. After that we will update
        .gantry_artifact.jsonl with the latest dataset info and create a new commit. HEAD file
        will also be updated using the latest commit information.

        Args:
            message (str): commit message, can't be empty
        Returns:
            Commit info
        """
        diff = self._get_diff()

        if (
            not diff[constants.NEW_FILES]
            and not diff[constants.MODIFIED_FILES]
            and not diff[constants.DELETED_FILES]
        ):
            logger.warning("No local changes to commit!")
            return self._get_current_commit()
        diff[constants.NEW_FILES] = self._upload_files(diff[constants.NEW_FILES])
        diff[constants.MODIFIED_FILES] = self._upload_files(diff[constants.MODIFIED_FILES])

        self._backup_file(constants.DATASET_MANIFEST_FILE)
        self._update_manifest_file(diff)

        result = self._upload_files(
            [
                DatasetFileInfo(
                    file_name=constants.DATASET_MANIFEST_FILE,
                    url=self._get_s3_url(constants.DATASET_MANIFEST_FILE),
                )
            ]
        )
        version_id = result[0].version_id
        try:
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/commits",
                json={
                    "message": message,
                    "metadata_s3_file_version": version_id,
                    "parent_commit_id": self._get_current_commit()["id"],
                    "email": self.user_email,
                },
                raise_for_status=True,
            )
            os.remove(
                self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.BACKUP_SUFFIX}")
            )
        except GantryRequestException as e:
            self._recover_file(constants.DATASET_MANIFEST_FILE)
            raise e

        commit_info = resp["data"]
        self._update_head(commit_info)

        return commit_info

    def sync(self, commit_id: Optional[str] = None) -> Dict[str, str]:
        """
        Sync local dataset folder based on a commit id. If commit_id was not provided, will sync
        based on the latest commit.

        Args:
            commit_id (str, optional): target commit id. Defaults to None.
        Returns:
            commit info
        """
        target_commit = (
            self.get_commit(commit_id=commit_id) if commit_id else self.get_latest_commit()
        )

        version_id = target_commit[constants.METADATA_S3_FILE_VERSION]

        self._download_files(
            [
                DatasetFileInfo(
                    file_name=constants.DATASET_MANIFEST_FILE,
                    url=self._get_s3_url(constants.DATASET_MANIFEST_FILE),
                    version_id=version_id,
                )
            ]
        )

        diff = self._get_diff(return_local_info=False)
        self._download_files(diff[constants.MODIFIED_FILES])  # Overwrite modified files
        self._download_files(diff[constants.DELETED_FILES])  # redownload deleted files

        # delete new added files
        for f in diff[constants.NEW_FILES]:
            os.remove(self._get_file_path(f.file_name))

        self._update_head(target_commit)

        return target_commit

    def get_commit(self, commit_id: str) -> Dict[str, str]:
        """
        Get commit details

        Args:
            commit_id (str): commit id
        Returns:
            Commit information
        """
        resp = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits/{commit_id}", raise_for_status=True
        )
        return resp["data"]

    def get_latest_commit(self) -> Dict[str, str]:
        """
        Get latest commit details

        Returns:
            Commit information
        """
        resp = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits", raise_for_status=True
        )
        return resp["data"][0]  # return latest commit

    def _update_manifest_file(self, diff: Dict[str, List[DatasetFileInfo]]):
        new_files = dict([(f.file_name, f) for f in diff[constants.NEW_FILES]])
        modified_files = dict([(f.file_name, f) for f in diff[constants.MODIFIED_FILES]])
        deleted_files = dict([(f.file_name, f) for f in diff[constants.DELETED_FILES]])

        with open(
            self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.NEW_SUFFIX}"), "w"
        ) as new_gantry_manifest:
            with open(
                self._get_file_path(constants.DATASET_MANIFEST_FILE), "r"
            ) as cur_gantry_manifest:
                for line in cur_gantry_manifest:
                    cur_file_info = json.loads(line)
                    cur_file_name = cur_file_info[constants.FILE_NAME]

                    if cur_file_name in modified_files:  # update modified files
                        new_gantry_manifest.write(modified_files[cur_file_name].to_jsonl())
                    elif cur_file_name in deleted_files:  # remove deleted files
                        continue
                    else:  # keep unchanged files
                        new_gantry_manifest.write(line)

            for _, file_info in new_files.items():
                new_gantry_manifest.write(file_info.to_jsonl())  # add new added files

        os.replace(
            self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.NEW_SUFFIX}"),
            self._get_file_path(constants.DATASET_MANIFEST_FILE),
        )

    def _get_diff(self, return_local_info=True):
        """
        Generate local diff.

        Args:
            return_local_info (bool): if true return local file info for modiflied files else
            return file info in dataset repo

        Returns:
            {
                "new_files": List[DatasetFileInfo],
                "modified_files": List[DatasetFileInfo],
                "deleted_files": List[DatasetFileInfo],
            }
        """
        repo_diff = defaultdict(list)

        commit_snapshot = {}
        with open(self._get_file_path(constants.DATASET_MANIFEST_FILE)) as cur_gantry_manifest:
            for line in cur_gantry_manifest:
                file_info = json.loads(line)
                commit_snapshot[file_info[constants.FILE_NAME]] = file_info

        local_files = get_files_checksum(Path(self.workspace) / self.dataset_name)

        # TODO:// Simplify the following logic to make it easier to maintain
        for file_name, checksum in local_files.items():
            if file_name.startswith(constants.GANTRY_FOLDER):  # skip .dataset_metadata folder
                continue
            elif file_name not in commit_snapshot:  # new added file
                repo_diff[constants.NEW_FILES].append(
                    DatasetFileInfo(
                        file_name=file_name,
                        url=self._get_s3_url(file_name),
                        sha256=checksum,
                    )
                )
            elif checksum != commit_snapshot[file_name].get(
                constants.SHA256
            ):  # if file_name in current_snapshot --> check if it has been modified
                if return_local_info:
                    repo_diff[constants.MODIFIED_FILES].append(
                        DatasetFileInfo(
                            file_name=file_name,
                            url=self._get_s3_url(file_name),
                            sha256=checksum,
                        )
                    )
                else:
                    repo_diff[constants.MODIFIED_FILES].append(
                        DatasetFileInfo(
                            file_name=file_name,
                            url=commit_snapshot[file_name][constants.URL],
                            version_id=commit_snapshot[file_name][constants.VERSION_ID],
                            sha256=commit_snapshot[file_name][constants.SHA256],
                        )
                    )

        for file_name in commit_snapshot:
            if file_name not in local_files:
                repo_diff[constants.DELETED_FILES].append(
                    DatasetFileInfo(
                        file_name=file_name,
                        url=commit_snapshot[file_name][constants.URL],
                        sha256=commit_snapshot[file_name][constants.SHA256],
                        version_id=commit_snapshot[file_name][constants.VERSION_ID],
                    )
                )
        return repo_diff

    def _get_file_path(self, file_name: str) -> str:
        return os.path.join(self.workspace, self.dataset_name, file_name)

    def _get_s3_url(self, file_name: str) -> str:
        return f"s3://{self._bucket_name}/{self._dataset_s3_prefix}/{self.dataset_name}/{file_name}"

    def _get_obj_key(self, s3_url: str) -> str:
        _, obj_key = parse_s3_path(s3_url)
        return obj_key

    def _update_head(self, commit_info: Dict[str, str]):
        with open(self._get_file_path(constants.DATASET_HEAD_FILE), "w") as f:
            json.dump(commit_info, f)

    def _get_current_commit(self) -> Dict[str, str]:
        with open(self._get_file_path(constants.DATASET_HEAD_FILE), "r") as f:
            return json.load(f)

    def _backup_file(self, file_path: str):
        # make a copy of the original file
        shutil.copyfile(
            self._get_file_path(file_path),
            self._get_file_path(f"{file_path}{constants.BACKUP_SUFFIX}"),
        )

    def _recover_file(self, file_path: str):
        # recover the file from a backup copy
        os.replace(
            self._get_file_path(f"{file_path}{constants.BACKUP_SUFFIX}"),
            self._get_file_path(file_path),
        )

    def _download_files(self, file_list: List[DatasetFileInfo]):
        """
        download file to the specific path

        Args:
            file_list (List[DatasetFileInfo]):
        """
        for i in range(0, len(file_list), BATCH_SIZE):
            batch = file_list[i : i + BATCH_SIZE]
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/presign/getobject",
                json={
                    "obj_infos": [
                        {
                            constants.OBJ_KEY: self._get_obj_key(item.url),
                            constants.VERSION_ID: item.version_id,
                        }
                        for item in batch
                    ],
                },
                raise_for_status=True,
            )
            presigned_urls = resp["data"]

            for item in batch:
                local_path = self._get_file_path(item.file_name)
                presigned_url = presigned_urls.get(self._get_obj_key(item.url))
                download_file_from_url(presigned_url=presigned_url, local_path=Path(local_path))

    def _upload_files(self, file_list: List[DatasetFileInfo]):
        """
        download file to the specific path

        Args:
            file_list (List[DatasetFileInfo]):
        Returns:
            List[DatasetFileInfo]
        """
        for i in range(0, len(file_list), BATCH_SIZE):
            batch = file_list[i : i + BATCH_SIZE]
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/presign/putobject",
                json={"obj_keys": [self._get_obj_key(f.url) for f in batch]},
                raise_for_status=True,
            )
            presigned_urls = resp["data"]

            for item in batch:
                local_path = self._get_file_path(item.file_name)
                presigned_url = presigned_urls.get(self._get_obj_key(item.url))
                version_id = upload_file_to_url(
                    presigned_url=presigned_url, local_path=Path(local_path)
                )
                item.version_id = version_id
                logger.info(f"successfully upload: {item.file_name}")

        return file_list
